class ITypeHintingFactory:
    def make_param_provider(self):
        """
        :rtype: rope.base.oi.type_hinting.providers.interfaces.IParamProvider
        """
        raise NotImplementedError

    def make_return_provider(self):
        """
        :rtype: rope.base.oi.type_hinting.providers.interfaces.IReturnProvider
        """
        raise NotImplementedError

    def make_assignment_provider(self):
        """
        :rtype: rope.base.oi.type_hinting.providers.interfaces.IAssignmentProvider
        """
        raise NotImplementedError

    def make_resolver(self):
        """
        :rtype: rope.base.oi.type_hinting.resolvers.interfaces.IResolver
        """
        raise NotImplementedError



import unittest
from textwrap import dedent

import rope.base.exceptions
from rope.refactor import change_signature
from ropetest import testutils


class ChangeSignatureTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_normalizing_parameters_for_trivial_case(self):
        code = dedent("""\
            def a_func():
                pass
            a_func()""")
        self.mod.write(code)
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(code, self.mod.read())

    def test_normalizing_parameters_for_trivial_case2(self):
        code = dedent("""\
            def a_func(param):
                pass
            a_func(2)""")
        self.mod.write(code)
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(code, self.mod.read())

    def test_normalizing_parameters_for_unneeded_keyword(self):
        self.mod.write(dedent("""\
            def a_func(param):
                pass
            a_func(param=1)"""))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(
            dedent("""\
                def a_func(param):
                    pass
                a_func(1)"""),
            self.mod.read(),
        )

    def test_normalizing_parameters_for_unneeded_keyword_for_methods(self):
        code = dedent("""\
            class A(object):
                def a_func(self, param):
                    pass
            a_var = A()
            a_var.a_func(param=1)
        """)
        self.mod.write(code)
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        expected = dedent("""\
            class A(object):
                def a_func(self, param):
                    pass
            a_var = A()
            a_var.a_func(1)
        """)
        self.assertEqual(expected, self.mod.read())

    def test_normalizing_parameters_for_unsorted_keyword(self):
        self.mod.write(dedent("""\
            def a_func(p1, p2):
                pass
            a_func(p2=2, p1=1)"""))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(
            dedent("""\
                def a_func(p1, p2):
                    pass
                a_func(1, 2)"""),
            self.mod.read(),
        )

    def test_raising_exceptions_for_non_functions(self):
        self.mod.write("a_var = 10")
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            change_signature.ChangeSignature(
                self.project, self.mod, self.mod.read().index("a_var") + 1
            )

    def test_normalizing_parameters_for_args_parameter(self):
        self.mod.write(dedent("""\
            def a_func(*arg):
                pass
            a_func(1, 2)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(
            dedent("""\
                def a_func(*arg):
                    pass
                a_func(1, 2)
            """),
            self.mod.read(),
        )

    def test_normalizing_parameters_for_args_parameter_and_keywords(self):
        self.mod.write(dedent("""\
            def a_func(param, *args):
                pass
            a_func(*[1, 2, 3])
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(
            dedent("""\
                def a_func(param, *args):
                    pass
                a_func(*[1, 2, 3])
            """),
            self.mod.read(),
        )

    def test_normalizing_functions_from_other_modules(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def a_func(param):
                pass
        """))
        self.mod.write(dedent("""\
            import mod1
            mod1.a_func(param=1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, mod1, mod1.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(
            dedent("""\
                import mod1
                mod1.a_func(1)
            """),
            self.mod.read(),
        )

    def test_normalizing_parameters_for_keyword_parameters(self):
        self.mod.write(dedent("""\
            def a_func(p1, **kwds):
                pass
            a_func(p2=2, p1=1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(
            dedent("""\
                def a_func(p1, **kwds):
                    pass
                a_func(1, p2=2)
            """),
            self.mod.read(),
        )

    def test_removing_arguments(self):
        self.mod.write(dedent("""\
            def a_func(p1):
                pass
            a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(0)]))
        self.assertEqual(
            dedent("""\
                def a_func():
                    pass
                a_func()
            """),
            self.mod.read(),
        )

    def test_removing_arguments_with_multiple_args(self):
        self.mod.write(dedent("""\
            def a_func(p1, p2):
                pass
            a_func(1, 2)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(0)]))
        self.assertEqual(
            dedent("""\
                def a_func(p2):
                    pass
                a_func(2)
            """),
            self.mod.read(),
        )

    def test_removing_arguments_passed_as_keywords(self):
        self.mod.write(dedent("""\
            def a_func(p1):
                pass
            a_func(p1=1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(0)]))
        self.assertEqual(
            dedent("""\
                def a_func():
                    pass
                a_func()
            """),
            self.mod.read(),
        )

    def test_removing_arguments_with_defaults(self):
        self.mod.write(dedent("""\
            def a_func(p1=1):
                pass
            a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(0)]))
        self.assertEqual(
            dedent("""\
                def a_func():
                    pass
                a_func()
            """),
            self.mod.read(),
        )

    def test_removing_arguments_star_args(self):
        self.mod.write(dedent("""\
            def a_func(p1, *args):
                pass
            a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(1)]))
        self.assertEqual(
            dedent("""\
                def a_func(p1):
                    pass
                a_func(1)
            """),
            self.mod.read(),
        )

    def test_removing_keyword_arg(self):
        self.mod.write(dedent("""\
            def a_func(p1, **kwds):
                pass
            a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(1)]))
        self.assertEqual(
            dedent("""\
                def a_func(p1):
                    pass
                a_func(1)
            """),
            self.mod.read(),
        )

    def test_removing_keyword_arg2(self):
        self.mod.write(dedent("""\
            def a_func(p1, *args, **kwds):
                pass
            a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(2)]))
        self.assertEqual(
            dedent("""\
                def a_func(p1, *args):
                    pass
                a_func(1)
            """),
            self.mod.read(),
        )

    # XXX: What to do here for star args?
    @unittest.skip("How to deal with start args?")
    def xxx_test_removing_arguments_star_args2(self):
        self.mod.write(dedent("""\
            def a_func(p1, *args):
                pass
            a_func(2, 3, p1=1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(1)]))
        self.assertEqual(
            dedent("""\
                def a_func(p1):
                    pass
                a_func(p1=1)
            """),
            self.mod.read(),
        )

    # XXX: What to do here for star args?
    def xxx_test_removing_arguments_star_args3(self):
        self.mod.write(dedent("""\
            def a_func(p1, *args):
                pass
            a_func(*[1, 2, 3])
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(1)]))
        self.assertEqual(
            dedent("""\
                def a_func(p1):
                    pass
                a_func(*[1, 2, 3])
            """),
            self.mod.read(),
        )

    def test_adding_arguments_for_normal_args_changing_definition(self):
        self.mod.write(dedent("""\
            def a_func():
                pass
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(
            signature.get_changes([change_signature.ArgumentAdder(0, "p1")])
        )
        self.assertEqual(
            dedent("""\
                def a_func(p1):
                    pass
            """),
            self.mod.read(),
        )

    def test_adding_arguments_for_normal_args_with_defaults(self):
        self.mod.write(dedent("""\
            def a_func():
                pass
            a_func()
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        adder = change_signature.ArgumentAdder(0, "p1", "None")
        self.project.do(signature.get_changes([adder]))
        self.assertEqual(
            dedent("""\
                def a_func(p1=None):
                    pass
                a_func()
            """),
            self.mod.read(),
        )

    def test_adding_arguments_for_normal_args_changing_calls(self):
        self.mod.write(dedent("""\
            def a_func():
                pass
            a_func()
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        adder = change_signature.ArgumentAdder(0, "p1", "None", "1")
        self.project.do(signature.get_changes([adder]))
        self.assertEqual(
            dedent("""\
                def a_func(p1=None):
                    pass
                a_func(1)
            """),
            self.mod.read(),
        )

    def test_adding_arguments_for_norm_args_chang_calls_with_kwords(self):
        self.mod.write(dedent("""\
            def a_func(p1=0):
                pass
            a_func()
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        adder = change_signature.ArgumentAdder(1, "p2", "0", "1")
        self.project.do(signature.get_changes([adder]))
        self.assertEqual(
            dedent("""\
                def a_func(p1=0, p2=0):
                    pass
                a_func(p2=1)
            """),
            self.mod.read(),
        )

    def test_adding_arguments_for_norm_args_chang_calls_with_no_value(self):
        self.mod.write(dedent("""\
            def a_func(p2=0):
                pass
            a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        adder = change_signature.ArgumentAdder(0, "p1", "0", None)
        self.project.do(signature.get_changes([adder]))
        self.assertEqual(
            dedent("""\
                def a_func(p1=0, p2=0):
                    pass
                a_func(p2=1)
            """),
            self.mod.read(),
        )

    def test_adding_duplicate_parameter_and_raising_exceptions(self):
        self.mod.write(dedent("""\
            def a_func(p1):
                pass
        """))
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            signature = change_signature.ChangeSignature(
                self.project, self.mod, self.mod.read().index("a_func") + 1
            )
            self.project.do(
                signature.get_changes([change_signature.ArgumentAdder(1, "p1")])
            )

    def test_inlining_default_arguments(self):
        self.mod.write(dedent("""\
            def a_func(p1=0):
                pass
            a_func()
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(
            signature.get_changes([change_signature.ArgumentDefaultInliner(0)])
        )
        self.assertEqual(
            dedent("""\
                def a_func(p1=0):
                    pass
                a_func(0)
            """),
            self.mod.read(),
        )

    def test_inlining_default_arguments2(self):
        self.mod.write(dedent("""\
            def a_func(p1=0):
                pass
            a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(
            signature.get_changes([change_signature.ArgumentDefaultInliner(0)])
        )
        self.assertEqual(
            dedent("""\
                def a_func(p1=0):
                    pass
                a_func(1)
            """),
            self.mod.read(),
        )

    def test_preserving_args_and_keywords_order(self):
        self.mod.write(dedent("""\
            def a_func(*args, **kwds):
                pass
            a_func(3, 1, 2, a=1, c=3, b=2)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentNormalizer()]))
        self.assertEqual(
            dedent("""\
                def a_func(*args, **kwds):
                    pass
                a_func(3, 1, 2, a=1, c=3, b=2)
            """),
            self.mod.read(),
        )

    def test_change_order_for_only_one_parameter(self):
        self.mod.write(dedent("""\
            def a_func(p1):
                pass
            a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(
            signature.get_changes([change_signature.ArgumentReorderer([0])])
        )
        self.assertEqual(
            dedent("""\
                def a_func(p1):
                    pass
                a_func(1)
            """),
            self.mod.read(),
        )

    def test_change_order_for_two_parameter(self):
        self.mod.write(dedent("""\
            def a_func(p1, p2):
                pass
            a_func(1, 2)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(
            signature.get_changes([change_signature.ArgumentReorderer([1, 0])])
        )
        self.assertEqual(
            dedent("""\
                def a_func(p2, p1):
                    pass
                a_func(2, 1)
            """),
            self.mod.read(),
        )

    def test_reordering_multi_line_function_headers(self):
        self.mod.write(dedent("""\
            def a_func(p1,
                p2):
                pass
            a_func(1, 2)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(
            signature.get_changes([change_signature.ArgumentReorderer([1, 0])])
        )
        self.assertEqual(
            dedent("""\
                def a_func(p2, p1):
                    pass
                a_func(2, 1)
            """),
            self.mod.read(),
        )

    def test_changing_order_with_static_params(self):
        self.mod.write(dedent("""\
            def a_func(p1, p2=0, p3=0):
                pass
            a_func(1, 2)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(
            signature.get_changes([change_signature.ArgumentReorderer([0, 2, 1])])
        )
        self.assertEqual(
            dedent("""\
                def a_func(p1, p3=0, p2=0):
                    pass
                a_func(1, p2=2)
            """),
            self.mod.read(),
        )

    def test_doing_multiple_changes(self):
        changers = []
        self.mod.write(dedent("""\
            def a_func(p1):
                pass
            a_func(1)
        """))
        changers.append(change_signature.ArgumentRemover(0))
        changers.append(change_signature.ArgumentAdder(0, "p2", None, None))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        signature.get_changes(changers).do()
        self.assertEqual(
            dedent("""\
                def a_func(p2):
                    pass
                a_func()
            """),
            self.mod.read(),
        )

    def test_doing_multiple_changes2(self):
        changers = []
        self.mod.write(dedent("""\
            def a_func(p1, p2):
                pass
            a_func(p2=2)
        """))
        changers.append(change_signature.ArgumentAdder(2, "p3", None, "3"))
        changers.append(change_signature.ArgumentReorderer([1, 0, 2]))
        changers.append(change_signature.ArgumentRemover(1))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        signature.get_changes(changers).do()
        self.assertEqual(
            dedent("""\
                def a_func(p2, p3):
                    pass
                a_func(2, 3)
            """),
            self.mod.read(),
        )

    def test_changing_signature_in_subclasses(self):
        self.mod.write(dedent("""\
            class A(object):
                def a_method(self):
                    pass
            class B(A):
                def a_method(self):
                    pass
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_method") + 1
        )
        signature.get_changes(
            [change_signature.ArgumentAdder(1, "p1")], in_hierarchy=True
        ).do()
        self.assertEqual(
            dedent("""\
                class A(object):
                    def a_method(self, p1):
                        pass
                class B(A):
                    def a_method(self, p1):
                        pass
            """),
            self.mod.read(),
        )

    def test_differentiating_class_accesses_from_instance_accesses(self):
        self.mod.write(dedent("""\
            class A(object):
                def a_func(self, param):
                    pass
            a_var = A()
            A.a_func(a_var, param=1)"""))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("a_func") + 1
        )
        self.project.do(signature.get_changes([change_signature.ArgumentRemover(1)]))
        self.assertEqual(
            dedent("""\
                class A(object):
                    def a_func(self):
                        pass
                a_var = A()
                A.a_func(a_var)"""),
            self.mod.read(),
        )

    def test_changing_signature_for_constructors(self):
        self.mod.write(dedent("""\
            class C(object):
                def __init__(self, p):
                    pass
            c = C(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("C") + 1
        )
        signature.get_changes([change_signature.ArgumentRemover(1)]).do()
        self.assertEqual(
            dedent("""\
                class C(object):
                    def __init__(self):
                        pass
                c = C()
            """),
            self.mod.read(),
        )

    def test_changing_signature_for_constructors2(self):
        self.mod.write(dedent("""\
            class C(object):
                def __init__(self, p):
                    pass
            c = C(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("__init__") + 1
        )
        signature.get_changes([change_signature.ArgumentRemover(1)]).do()
        self.assertEqual(
            dedent("""\
                class C(object):
                    def __init__(self):
                        pass
                c = C()
            """),
            self.mod.read(),
        )

    def test_changing_signature_for_constructors_when_using_super(self):
        self.mod.write(dedent("""\
            class A(object):
                def __init__(self, p):
                    pass
            class B(A):
                def __init__(self, p):
                    super(B, self).__init__(p)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().index("__init__") + 1
        )
        signature.get_changes([change_signature.ArgumentRemover(1)]).do()
        self.assertEqual(
            dedent("""\
                class A(object):
                    def __init__(self):
                        pass
                class B(A):
                    def __init__(self, p):
                        super(B, self).__init__()
            """),
            self.mod.read(),
        )

    def test_redordering_arguments_reported_by_mft(self):
        self.mod.write(dedent("""\
            def f(a, b, c):
                pass
            f(1, 2, 3)
        """))
        signature = change_signature.ChangeSignature(
            self.project, self.mod, self.mod.read().rindex("f")
        )
        signature.get_changes([change_signature.ArgumentReorderer([1, 2, 0])]).do()
        self.assertEqual(
            dedent("""\
                def f(b, c, a):
                    pass
                f(2, 3, 1)
            """),
            self.mod.read(),
        )

    def test_resources_parameter(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write("def a_func(param):\n    pass\n")
        self.mod.write(dedent("""\
            import mod1
            mod1.a_func(1)
        """))
        signature = change_signature.ChangeSignature(
            self.project, mod1, mod1.read().index("a_func") + 1
        )
        signature.get_changes(
            [change_signature.ArgumentRemover(0)], resources=[mod1]
        ).do()
        self.assertEqual(
            dedent("""\
                import mod1
                mod1.a_func(1)
            """),
            self.mod.read(),
        )
        self.assertEqual(
            dedent("""\
                def a_func():
                    pass
            """),
            mod1.read(),
        )

    def test_reordering_and_automatic_defaults(self):
        code = dedent("""\
            def f(p1, p2=2):
                pass
            f(1, 2)
        """)
        self.mod.write(code)
        signature = change_signature.ChangeSignature(
            self.project, self.mod, code.index("f(")
        )
        reorder = change_signature.ArgumentReorderer([1, 0], autodef="1")
        signature.get_changes([reorder]).do()
        expected = dedent("""\
            def f(p2=2, p1=1):
                pass
            f(2, 1)
        """)
        self.assertEqual(expected, self.mod.read())



import os
import unittest
from textwrap import dedent

from rope.base import exceptions
from ropetest import testutils


class PythonFileRunnerTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def make_sample_python_file(self, file_path, get_text_function_source=None):
        self.project.root.create_file(file_path)
        file = self.project.get_resource(file_path)
        if not get_text_function_source:
            get_text_function_source = "def get_text():\n    return 'run'\n\n"
        file_content = (
            get_text_function_source + "output = open('output.txt', 'w')\n"
            "output.write(get_text())\noutput.close()\n"
        )
        file.write(file_content)

    def get_output_file_content(self, file_path):
        try:
            output_path = ""
            last_slash = file_path.rfind("/")
            if last_slash != -1:
                output_path = file_path[0 : last_slash + 1]
            file = self.project.get_resource(output_path + "output.txt")
            return file.read()
        except exceptions.ResourceNotFoundError:
            return ""

    def test_making_runner(self):
        file_path = "sample.py"
        self.make_sample_python_file(file_path)
        file_resource = self.project.get_resource(file_path)
        runner = self.pycore.run_module(file_resource)
        runner.wait_process()
        self.assertEqual("run", self.get_output_file_content(file_path))

    def test_passing_arguments(self):
        file_path = "sample.py"
        function_source = dedent("""\
            import sys
            def get_text():
                return str(sys.argv[1:])
        """)
        self.make_sample_python_file(file_path, function_source)
        file_resource = self.project.get_resource(file_path)
        runner = self.pycore.run_module(file_resource, args=["hello", "world"])
        runner.wait_process()
        self.assertTrue(
            self.get_output_file_content(file_path).endswith("['hello', 'world']")
        )

    def test_passing_arguments_with_spaces(self):
        file_path = "sample.py"
        function_source = dedent("""\
            import sys
            def get_text():
                return str(sys.argv[1:])
        """)
        self.make_sample_python_file(file_path, function_source)
        file_resource = self.project.get_resource(file_path)
        runner = self.pycore.run_module(file_resource, args=["hello world"])
        runner.wait_process()
        self.assertTrue(
            self.get_output_file_content(file_path).endswith("['hello world']")
        )

    def test_killing_runner(self):
        file_path = "sample.py"
        code = dedent("""\
            def get_text():
                import time
                time.sleep(1)
                return 'run'
        """)
        self.make_sample_python_file(
            file_path,
            code,
        )
        file_resource = self.project.get_resource(file_path)
        runner = self.pycore.run_module(file_resource)
        runner.kill_process()
        self.assertEqual("", self.get_output_file_content(file_path))

    def test_running_nested_files(self):
        self.project.root.create_folder("src")
        file_path = "src/sample.py"
        self.make_sample_python_file(file_path)
        file_resource = self.project.get_resource(file_path)
        runner = self.pycore.run_module(file_resource)
        runner.wait_process()
        self.assertEqual("run", self.get_output_file_content(file_path))

    def test_setting_process_input(self):
        file_path = "sample.py"
        code = dedent("""\
            def get_text():
                import sys
                return sys.stdin.readline()
        """)
        self.make_sample_python_file(file_path, code)
        temp_file_name = "processtest.tmp"
        try:
            temp_file = open(temp_file_name, "w")
            temp_file.write("input text\n")
            temp_file.close()
            file_resource = self.project.get_resource(file_path)
            stdin = open(temp_file_name)
            runner = self.pycore.run_module(file_resource, stdin=stdin)
            runner.wait_process()
            stdin.close()
            self.assertEqual("input text\n", self.get_output_file_content(file_path))
        finally:
            os.remove(temp_file_name)

    def test_setting_process_output(self):
        file_path = "sample.py"
        code = dedent("""\
            def get_text():
                print('output text')
                return 'run'
        """)
        self.make_sample_python_file(file_path, code)
        temp_file_name = "processtest.tmp"
        try:
            file_resource = self.project.get_resource(file_path)
            stdout = open(temp_file_name, "w")
            runner = self.pycore.run_module(file_resource, stdout=stdout)
            runner.wait_process()
            stdout.close()
            temp_file = open(temp_file_name)
            self.assertEqual("output text\n", temp_file.read())
            temp_file.close()
        finally:
            os.remove(temp_file_name)

    def test_setting_pythonpath(self):
        src = self.project.root.create_folder("src")
        src.create_file("sample.py")
        src.get_child("sample.py").write("def f():\n    pass\n")
        self.project.root.create_folder("test")
        file_path = "test/test.py"
        code = dedent("""\
            def get_text():
                import sample
                sample.f()
                return'run'
        """)
        self.make_sample_python_file(file_path, code)
        file_resource = self.project.get_resource(file_path)
        runner = self.pycore.run_module(file_resource)
        runner.wait_process()
        self.assertEqual("run", self.get_output_file_content(file_path))

    def test_making_runner_when_doi_is_disabled(self):
        self.project.set("enable_doi", False)
        file_path = "sample.py"
        self.make_sample_python_file(file_path)
        file_resource = self.project.get_resource(file_path)
        runner = self.pycore.run_module(file_resource)
        runner.wait_process()
        self.assertEqual("run", self.get_output_file_content(file_path))



import sys
import unittest
from textwrap import dedent

import rope.base.codeanalyze
import rope.refactor.occurrences
from rope.refactor import rename
from rope.refactor.rename import Rename
from ropetest import testutils


class RenameRefactoringTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _local_rename(self, source_code, offset, new_name):
        testmod = testutils.create_module(self.project, "testmod")
        testmod.write(source_code)
        changes = Rename(self.project, testmod, offset).get_changes(
            new_name, resources=[testmod]
        )
        self.project.do(changes)
        return testmod.read()

    def _rename(self, resource, offset, new_name, **kwds):
        changes = Rename(self.project, resource, offset).get_changes(new_name, **kwds)
        self.project.do(changes)

    def test_local_variable_but_not_parameter(self):
        code = dedent("""\
            a = 10
            foo = dict(a=a)
        """)

        refactored = self._local_rename(code, 1, "new_a")
        self.assertEqual(
            dedent("""\
                new_a = 10
                foo = dict(a=new_a)
            """),
            refactored,
        )

    def test_simple_global_variable_renaming(self):
        refactored = self._local_rename("a_var = 20\n", 2, "new_var")
        self.assertEqual("new_var = 20\n", refactored)

    def test_variable_renaming_only_in_its_scope(self):
        refactored = self._local_rename(
            dedent("""\
                a_var = 20
                def a_func():
                    a_var = 10
            """),
            32,
            "new_var",
        )
        self.assertEqual(
            dedent("""\
                a_var = 20
                def a_func():
                    new_var = 10
            """),
            refactored,
        )

    def test_not_renaming_dot_name(self):
        refactored = self._local_rename(
            dedent("""\
                replace = True
                'aaa'.replace('a', 'b')
            """),
            1,
            "new_var",
        )
        self.assertEqual(
            dedent("""\
                new_var = True
                'aaa'.replace('a', 'b')
            """),
            refactored,
        )

    def test_renaming_multiple_names_in_the_same_line(self):
        refactored = self._local_rename(
            dedent("""\
                a_var = 10
                a_var = 10 + a_var / 2
            """),
            2,
            "new_var",
        )
        self.assertEqual(
            dedent("""\
                new_var = 10
                new_var = 10 + new_var / 2
            """),
            refactored,
        )

    def test_renaming_names_when_getting_some_attribute(self):
        refactored = self._local_rename(
            dedent("""\
                a_var = 'a b c'
                a_var.split('\\n')
            """),
            2,
            "new_var",
        )
        self.assertEqual(
            dedent("""\
                new_var = 'a b c'
                new_var.split('\\n')
            """),
            refactored,
        )

    def test_renaming_names_when_getting_some_attribute2(self):
        refactored = self._local_rename(
            dedent("""\
                a_var = 'a b c'
                a_var.split('\\n')
            """),
            20,
            "new_var",
        )
        self.assertEqual(
            dedent("""\
                new_var = 'a b c'
                new_var.split('\\n')
            """),
            refactored,
        )

    def test_renaming_function_parameters1(self):
        refactored = self._local_rename(
            dedent("""\
                def f(a_param):
                    print(a_param)
            """),
            8,
            "new_param",
        )
        self.assertEqual(
            dedent("""\
                def f(new_param):
                    print(new_param)
            """),
            refactored,
        )

    def test_renaming_function_parameters2(self):
        refactored = self._local_rename(
            dedent("""\
                def f(a_param):
                    print(a_param)
            """),
            30,
            "new_param",
        )
        self.assertEqual(
            dedent("""\
                def f(new_param):
                    print(new_param)
            """),
            refactored,
        )

    def test_renaming_occurrences_inside_functions(self):
        code = dedent("""\
            def a_func(p1):
                a = p1
            a_func(1)
        """)
        refactored = self._local_rename(code, code.index("p1") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param):
                    a = new_param
                a_func(1)
            """),
            refactored,
        )

    def test_renaming_comprehension_loop_variables(self):
        code = "[b_var for b_var, c_var in d_var if b_var == c_var]"
        refactored = self._local_rename(code, code.index("b_var") + 1, "new_var")
        self.assertEqual(
            "[new_var for new_var, c_var in d_var if new_var == c_var]", refactored
        )

    def test_renaming_list_comprehension_loop_variables_in_assignment(self):
        code = "a_var = [b_var for b_var, c_var in d_var if b_var == c_var]"
        refactored = self._local_rename(code, code.index("b_var") + 1, "new_var")
        self.assertEqual(
            "a_var = [new_var for new_var, c_var in d_var if new_var == c_var]",
            refactored,
        )

    def test_renaming_generator_comprehension_loop_variables(self):
        code = "a_var = (b_var for b_var, c_var in d_var if b_var == c_var)"
        refactored = self._local_rename(code, code.index("b_var") + 1, "new_var")
        self.assertEqual(
            "a_var = (new_var for new_var, c_var in d_var if new_var == c_var)",
            refactored,
        )

    def test_renaming_comprehension_loop_variables_scope(self):
        code = dedent("""\
            [b_var for b_var, c_var in d_var if b_var == c_var]
            b_var = 10
        """)
        refactored = self._local_rename(code, code.index("b_var") + 1, "new_var")
        self.assertEqual(
            dedent("""\
                [new_var for new_var, c_var in d_var if new_var == c_var]
                b_var = 10
            """),
            refactored,
        )

    @testutils.only_for_versions_higher("3.8")
    def test_renaming_inline_assignment(self):
        code = dedent("""\
            while a_var := next(foo):
                print(a_var)
        """)
        refactored = self._local_rename(code, code.index("a_var") + 1, "new_var")
        self.assertEqual(
            dedent("""\
                while new_var := next(foo):
                    print(new_var)
            """),
            refactored,
        )

    def test_renaming_arguments_for_normal_args_changing_calls(self):
        code = dedent("""\
            def a_func(p1=None, p2=None):
                pass
            a_func(p2=1)
        """)
        refactored = self._local_rename(code, code.index("p2") + 1, "p3")
        self.assertEqual(
            dedent("""\
                def a_func(p1=None, p3=None):
                    pass
                a_func(p3=1)
            """),
            refactored,
        )

    def test_renaming_function_parameters_of_class_init(self):
        code = dedent("""\
            class A(object):
                def __init__(self, a_param):
                    pass
            a_var = A(a_param=1)
        """)
        refactored = self._local_rename(code, code.index("a_param") + 1, "new_param")
        expected = dedent("""\
            class A(object):
                def __init__(self, new_param):
                    pass
            a_var = A(new_param=1)
        """)
        self.assertEqual(expected, refactored)

    def test_rename_functions_parameters_and_occurrences_in_other_modules(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write(dedent("""\
            def a_func(a_param):
                print(a_param)
        """))
        mod2.write(dedent("""\
            from mod1 import a_func
            a_func(a_param=10)
        """))
        self._rename(mod1, mod1.read().index("a_param") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param):
                    print(new_param)
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                from mod1 import a_func
                a_func(new_param=10)
            """),
            mod2.read(),
        )

    def test_renaming_with_backslash_continued_names(self):
        refactored = self._local_rename(
            "replace = True\n'ali'.\\\nreplace\n", 2, "is_replace"
        )
        self.assertEqual("is_replace = True\n'ali'.\\\nreplace\n", refactored)

    def test_renaming_occurrence_in_f_string(self):
        code = dedent("""\
            a_var = 20
            a_string=f'value: {a_var}'
        """)
        expected = dedent("""\
            new_var = 20
            a_string=f'value: {new_var}'
        """)
        refactored = self._local_rename(code, 2, "new_var")
        self.assertEqual(expected, refactored)

    def test_renaming_occurrence_in_nested_f_string(self):
        code = dedent("""\
            a_var = 20
            a_string=f'{f"{a_var}"}'
        """)
        expected = dedent("""\
            new_var = 20
            a_string=f'{f"{new_var}"}'
        """)
        refactored = self._local_rename(code, 2, "new_var")
        self.assertEqual(expected, refactored)

    def test_not_renaming_string_contents_in_f_string(self):
        refactored = self._local_rename(
            "a_var = 20\na_string=f'{\"a_var\"}'\n", 2, "new_var"
        )
        self.assertEqual(
            dedent("""\
                new_var = 20
                a_string=f'{"a_var"}'
            """),
            refactored,
        )

    def test_not_renaming_string_contents(self):
        refactored = self._local_rename("a_var = 20\na_string='a_var'\n", 2, "new_var")
        self.assertEqual(
            dedent("""\
                new_var = 20
                a_string='a_var'
            """),
            refactored,
        )

    def test_not_renaming_comment_contents(self):
        refactored = self._local_rename("a_var = 20\n# a_var\n", 2, "new_var")
        self.assertEqual(
            dedent("""\
                new_var = 20
                # a_var
            """),
            refactored,
        )

    def test_renaming_all_occurrences_in_containing_scope(self):
        code = dedent("""\
            if True:
                a_var = 1
            else:
                a_var = 20
        """)
        refactored = self._local_rename(code, 16, "new_var")
        self.assertEqual(
            dedent("""\
                if True:
                    new_var = 1
                else:
                    new_var = 20
            """),
            refactored,
        )

    def test_renaming_a_variable_with_arguement_name(self):
        code = dedent("""\
            a_var = 10
            def a_func(a_var):
                print(a_var)
        """)
        refactored = self._local_rename(code, 1, "new_var")
        self.assertEqual(
            dedent("""\
                new_var = 10
                def a_func(a_var):
                    print(a_var)
            """),
            refactored,
        )

    def test_renaming_an_arguement_with_variable_name(self):
        code = dedent("""\
            a_var = 10
            def a_func(a_var):
                print(a_var)
        """)
        refactored = self._local_rename(code, len(code) - 3, "new_var")
        self.assertEqual(
            dedent("""\
                a_var = 10
                def a_func(new_var):
                    print(new_var)
            """),
            refactored,
        )

    def test_renaming_function_with_local_variable_name(self):
        code = dedent("""\
            def a_func():
                a_func=20
            a_func()""")
        refactored = self._local_rename(code, len(code) - 3, "new_func")
        self.assertEqual(
            dedent("""\
                def new_func():
                    a_func=20
                new_func()"""),
            refactored,
        )

    def test_renaming_functions(self):
        code = dedent("""\
            def a_func():
                pass
            a_func()
        """)
        refactored = self._local_rename(code, len(code) - 5, "new_func")
        self.assertEqual(
            dedent("""\
                def new_func():
                    pass
                new_func()
            """),
            refactored,
        )

    def test_renaming_async_function(self):
        code = dedent("""\
            async def a_func():
                pass
            a_func()""")
        refactored = self._local_rename(code, len(code) - 5, "new_func")
        self.assertEqual(
            dedent("""\
                async def new_func():
                    pass
                new_func()"""),
            refactored,
        )

    def test_renaming_await(self):
        code = dedent("""\
            async def b_func():
                pass
            async def a_func():
                await b_func()""")
        refactored = self._local_rename(code, len(code) - 5, "new_func")
        self.assertEqual(
            dedent("""\
                async def new_func():
                    pass
                async def a_func():
                    await new_func()"""),
            refactored,
        )

    def test_renaming_functions_across_modules(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def a_func():
                pass
            a_func()
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            import mod1
            mod1.a_func()
        """))
        self._rename(mod1, len(mod1.read()) - 5, "new_func")
        self.assertEqual(
            dedent("""\
                def new_func():
                    pass
                new_func()
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                import mod1
                mod1.new_func()
            """),
            mod2.read(),
        )

    def test_renaming_functions_across_modules_from_import(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def a_func():
                pass
            a_func()
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            from mod1 import a_func
            a_func()
        """))
        self._rename(mod1, len(mod1.read()) - 5, "new_func")
        self.assertEqual(
            dedent("""\
                def new_func():
                    pass
                new_func()
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                from mod1 import new_func
                new_func()
            """),
            mod2.read(),
        )

    def test_renaming_functions_from_another_module(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def a_func():
                pass
            a_func()
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            import mod1
            mod1.a_func()
        """))
        self._rename(mod2, len(mod2.read()) - 5, "new_func")
        self.assertEqual(
            dedent("""\
                def new_func():
                    pass
                new_func()
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                import mod1
                mod1.new_func()
            """),
            mod2.read(),
        )

    def test_applying_all_changes_together(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            import mod2
            mod2.a_func()
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            def a_func():
                pass
            a_func()
        """))
        self._rename(mod2, len(mod2.read()) - 5, "new_func")
        self.assertEqual(
            dedent("""\
                import mod2
                mod2.new_func()
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                def new_func():
                    pass
                new_func()
            """),
            mod2.read(),
        )

    def test_renaming_modules(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def a_func():
                pass
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write("from mod1 import a_func\n")
        self._rename(mod2, mod2.read().index("mod1") + 1, "newmod")
        self.assertTrue(
            not mod1.exists() and self.project.find_module("newmod") is not None
        )
        self.assertEqual("from newmod import a_func\n", mod2.read())

    def test_renaming_modules_aliased(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def a_func():
                pass
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            import mod1 as m
            m.a_func()
        """))
        self._rename(mod1, None, "newmod")
        self.assertTrue(
            not mod1.exists() and self.project.find_module("newmod") is not None
        )
        self.assertEqual("import newmod as m\nm.a_func()\n", mod2.read())

    def test_renaming_packages(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1", pkg)
        mod1.write(dedent("""\
            def a_func():
                pass
        """))
        mod2 = testutils.create_module(self.project, "mod2", pkg)
        mod2.write("from pkg.mod1 import a_func\n")
        self._rename(mod2, 6, "newpkg")
        self.assertTrue(self.project.find_module("newpkg.mod1") is not None)
        new_mod2 = self.project.find_module("newpkg.mod2")
        self.assertEqual("from newpkg.mod1 import a_func\n", new_mod2.read())

    def test_module_dependencies(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            class AClass(object):
                pass
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            import mod1
            a_var = mod1.AClass()
        """))
        self.project.get_pymodule(mod2).get_attributes()["mod1"]
        mod1.write(dedent("""\
            def AClass():
                return 0
        """))

        self._rename(mod2, len(mod2.read()) - 3, "a_func")
        self.assertEqual(
            dedent("""\
                def a_func():
                    return 0
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                import mod1
                a_var = mod1.a_func()
            """),
            mod2.read(),
        )

    def test_renaming_class_attributes(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            class AClass(object):
                def __init__(self):
                    self.an_attr = 10
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            import mod1
            a_var = mod1.AClass()
            another_var = a_var.an_attr"""))

        self._rename(mod1, mod1.read().index("an_attr"), "attr")
        self.assertEqual(
            dedent("""\
                class AClass(object):
                    def __init__(self):
                        self.attr = 10
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                import mod1
                a_var = mod1.AClass()
                another_var = a_var.attr"""),
            mod2.read(),
        )

    def test_renaming_class_attributes2(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            class AClass(object):
                def __init__(self):
                    an_attr = 10
                    self.an_attr = 10
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            import mod1
            a_var = mod1.AClass()
            another_var = a_var.an_attr"""))

        self._rename(mod1, mod1.read().rindex("an_attr"), "attr")
        self.assertEqual(
            dedent("""\
                class AClass(object):
                    def __init__(self):
                        an_attr = 10
                        self.attr = 10
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                import mod1
                a_var = mod1.AClass()
                another_var = a_var.attr"""),
            mod2.read(),
        )

    def test_renaming_methods_in_subclasses(self):
        mod = testutils.create_module(self.project, "mod1")
        mod.write(dedent("""\
            class A(object):
                def a_method(self):
                    pass
            class B(A):
                def a_method(self):
                    pass
        """))

        self._rename(
            mod, mod.read().rindex("a_method") + 1, "new_method", in_hierarchy=True
        )
        self.assertEqual(
            dedent("""\
                class A(object):
                    def new_method(self):
                        pass
                class B(A):
                    def new_method(self):
                        pass
            """),
            mod.read(),
        )

    def test_renaming_methods_in_sibling_classes(self):
        mod = testutils.create_module(self.project, "mod1")
        mod.write(dedent("""\
            class A(object):
                def a_method(self):
                    pass
            class B(A):
                def a_method(self):
                    pass
            class C(A):
                def a_method(self):
                    pass
        """))

        self._rename(
            mod, mod.read().rindex("a_method") + 1, "new_method", in_hierarchy=True
        )
        self.assertEqual(
            dedent("""\
                class A(object):
                    def new_method(self):
                        pass
                class B(A):
                    def new_method(self):
                        pass
                class C(A):
                    def new_method(self):
                        pass
            """),
            mod.read(),
        )

    def test_not_renaming_methods_in_hierarchies(self):
        mod = testutils.create_module(self.project, "mod1")
        mod.write(dedent("""\
            class A(object):
                def a_method(self):
                    pass
            class B(A):
                def a_method(self):
                    pass
        """))

        self._rename(
            mod, mod.read().rindex("a_method") + 1, "new_method", in_hierarchy=False
        )
        self.assertEqual(
            dedent("""\
                class A(object):
                    def a_method(self):
                        pass
                class B(A):
                    def new_method(self):
                        pass
            """),
            mod.read(),
        )

    def test_undoing_refactorings(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def a_func():
                pass
            a_func()
        """))
        self._rename(mod1, len(mod1.read()) - 5, "new_func")
        self.project.history.undo()
        self.assertEqual(
            dedent("""\
                def a_func():
                    pass
                a_func()
            """),
            mod1.read(),
        )

    def test_undoing_renaming_modules(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def a_func():
                pass
        """))
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write("from mod1 import a_func\n")
        self._rename(mod2, 6, "newmod")
        self.project.history.undo()
        self.assertEqual("mod1.py", mod1.path)
        self.assertEqual("from mod1 import a_func\n", mod2.read())

    def test_rename_in_module_renaming_one_letter_names_for_expressions(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write("a = 10\nprint(1+a)\n")
        pymod = self.project.get_module("mod1")
        old_pyname = pymod["a"]
        finder = rope.refactor.occurrences.create_finder(self.project, "a", old_pyname)
        refactored = rename.rename_in_module(
            finder, "new_var", pymodule=pymod, replace_primary=True
        )
        self.assertEqual(
            dedent("""\
                new_var = 10
                print(1+new_var)
            """),
            refactored,
        )

    def test_renaming_for_loop_variable(self):
        code = dedent("""\
            for var in range(10):
                print(var)
        """)
        refactored = self._local_rename(code, code.find("var") + 1, "new_var")
        self.assertEqual(
            dedent("""\
                for new_var in range(10):
                    print(new_var)
            """),
            refactored,
        )

    def test_renaming_async_for_loop_variable(self):
        code = dedent("""\
            async def func():
                async for var in range(10):
                    print(var)
        """)
        refactored = self._local_rename(code, code.find("var") + 1, "new_var")
        self.assertEqual(
            dedent("""\
                async def func():
                    async for new_var in range(10):
                        print(new_var)
            """),
            refactored,
        )

    def test_renaming_async_with_context_manager(self):
        code = dedent("""\
            def a_cm(): pass
            async def a_func():
                async with a_cm() as x: pass""")
        refactored = self._local_rename(code, code.find("a_cm") + 1, "another_cm")
        expected = dedent("""\
            def another_cm(): pass
            async def a_func():
                async with another_cm() as x: pass""")
        self.assertEqual(refactored, expected)

    def test_renaming_async_with_as_variable(self):
        code = dedent("""\
            async def func():
                async with a_func() as var:
                    print(var)
        """)
        refactored = self._local_rename(code, code.find("var") + 1, "new_var")
        self.assertEqual(
            dedent("""\
                async def func():
                    async with a_func() as new_var:
                        print(new_var)
            """),
            refactored,
        )

    def test_renaming_parameters(self):
        code = dedent("""\
            def a_func(param):
                print(param)
            a_func(param=hey)
        """)
        refactored = self._local_rename(code, code.find("param") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param):
                    print(new_param)
                a_func(new_param=hey)
            """),
            refactored,
        )

    def test_renaming_assigned_parameters(self):
        code = dedent("""\
            def f(p):
                p = p + 1
                return p
            f(p=1)
        """)
        refactored = self._local_rename(code, code.find("p"), "arg")
        self.assertEqual(
            dedent("""\
                def f(arg):
                    arg = arg + 1
                    return arg
                f(arg=1)
            """),
            refactored,
        )

    def test_renaming_parameters_not_renaming_others(self):
        code = dedent("""\
            def a_func(param):
                print(param)
            param=10
            a_func(param)
        """)
        refactored = self._local_rename(code, code.find("param") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param):
                    print(new_param)
                param=10
                a_func(param)
            """),
            refactored,
        )

    def test_renaming_parameters_not_renaming_others2(self):
        code = dedent("""\
            def a_func(param):
                print(param)
            param=10
            a_func(param=param)""")
        refactored = self._local_rename(code, code.find("param") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param):
                    print(new_param)
                param=10
                a_func(new_param=param)"""),
            refactored,
        )

    def test_renaming_parameters_with_multiple_params(self):
        code = dedent("""\
            def a_func(param1, param2):
                print(param1)
            a_func(param1=1, param2=2)
        """)
        refactored = self._local_rename(code, code.find("param1") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param, param2):
                    print(new_param)
                a_func(new_param=1, param2=2)
            """),
            refactored,
        )

    def test_renaming_parameters_with_multiple_params2(self):
        code = dedent("""\
            def a_func(param1, param2):
                print(param1)
            a_func(param1=1, param2=2)
        """)
        refactored = self._local_rename(code, code.rfind("param2") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(param1, new_param):
                    print(param1)
                a_func(param1=1, new_param=2)
            """),
            refactored,
        )

    def test_renaming_parameters_on_calls(self):
        code = dedent("""\
            def a_func(param):
                print(param)
            a_func(param = hey)
        """)
        refactored = self._local_rename(code, code.rfind("param") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param):
                    print(new_param)
                a_func(new_param = hey)
            """),
            refactored,
        )

    def test_renaming_parameters_spaces_before_call(self):
        code = dedent("""\
            def a_func(param):
                print(param)
            a_func  (param=hey)
        """)
        refactored = self._local_rename(code, code.rfind("param") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param):
                    print(new_param)
                a_func  (new_param=hey)
            """),
            refactored,
        )

    def test_renaming_parameter_like_objects_after_keywords(self):
        code = dedent("""\
            def a_func(param):
                print(param)
            dict(param=hey)
        """)
        refactored = self._local_rename(code, code.find("param") + 1, "new_param")
        self.assertEqual(
            dedent("""\
                def a_func(new_param):
                    print(new_param)
                dict(param=hey)
            """),
            refactored,
        )

    def test_renaming_variables_in_init_dot_pys(self):
        pkg = testutils.create_package(self.project, "pkg")
        init_dot_py = pkg.get_child("__init__.py")
        init_dot_py.write("a_var = 10\n")
        mod = testutils.create_module(self.project, "mod")
        mod.write("import pkg\nprint(pkg.a_var)\n")
        self._rename(mod, mod.read().index("a_var") + 1, "new_var")
        self.assertEqual("new_var = 10\n", init_dot_py.read())
        self.assertEqual("import pkg\nprint(pkg.new_var)\n", mod.read())

    def test_renaming_variables_in_init_dot_pys2(self):
        pkg = testutils.create_package(self.project, "pkg")
        init_dot_py = pkg.get_child("__init__.py")
        init_dot_py.write("a_var = 10\n")
        mod = testutils.create_module(self.project, "mod")
        mod.write("import pkg\nprint(pkg.a_var)\n")
        self._rename(init_dot_py, init_dot_py.read().index("a_var") + 1, "new_var")
        self.assertEqual("new_var = 10\n", init_dot_py.read())
        self.assertEqual("import pkg\nprint(pkg.new_var)\n", mod.read())

    def test_renaming_variables_in_init_dot_pys3(self):
        pkg = testutils.create_package(self.project, "pkg")
        init_dot_py = pkg.get_child("__init__.py")
        init_dot_py.write("a_var = 10\n")
        mod = testutils.create_module(self.project, "mod")
        mod.write("import pkg\nprint(pkg.a_var)\n")
        self._rename(mod, mod.read().index("a_var") + 1, "new_var")
        self.assertEqual("new_var = 10\n", init_dot_py.read())
        self.assertEqual("import pkg\nprint(pkg.new_var)\n", mod.read())

    def test_renaming_resources_using_rename_module_refactoring(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("a_var = 1")
        mod2.write("import mod1\nmy_var = mod1.a_var\n")
        renamer = rename.Rename(self.project, mod1)
        renamer.get_changes("newmod").do()
        self.assertEqual("import newmod\nmy_var = newmod.a_var\n", mod2.read())

    def test_renam_resources_using_rename_module_refactor_for_packages(self):
        mod1 = testutils.create_module(self.project, "mod1")
        pkg = testutils.create_package(self.project, "pkg")
        mod1.write("import pkg\nmy_pkg = pkg")
        renamer = rename.Rename(self.project, pkg)
        renamer.get_changes("newpkg").do()
        self.assertEqual("import newpkg\nmy_pkg = newpkg", mod1.read())

    def test_renam_resources_use_rename_module_refactor_for_init_dot_py(self):
        mod1 = testutils.create_module(self.project, "mod1")
        pkg = testutils.create_package(self.project, "pkg")
        mod1.write("import pkg\nmy_pkg = pkg")
        renamer = rename.Rename(self.project, pkg.get_child("__init__.py"))
        renamer.get_changes("newpkg").do()
        self.assertEqual("import newpkg\nmy_pkg = newpkg", mod1.read())

    def test_renaming_global_variables(self):
        code = dedent("""\
            a_var = 1
            def a_func():
                global a_var
                var = a_var
        """)
        refactored = self._local_rename(code, code.index("a_var"), "new_var")
        self.assertEqual(
            dedent("""\
                new_var = 1
                def a_func():
                    global new_var
                    var = new_var
            """),
            refactored,
        )

    def test_renaming_global_variables2(self):
        code = dedent("""\
            a_var = 1
            def a_func():
                global a_var
                var = a_var
        """)
        refactored = self._local_rename(code, code.rindex("a_var"), "new_var")
        self.assertEqual(
            dedent("""\
                new_var = 1
                def a_func():
                    global new_var
                    var = new_var
            """),
            refactored,
        )

    def test_renaming_when_unsure(self):
        code = dedent("""\
            class C(object):
                def a_func(self):
                    pass
            def f(arg):
                arg.a_func()
        """)
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(code)
        self._rename(mod1, code.index("a_func"), "new_func", unsure=self._true)
        self.assertEqual(
            dedent("""\
                class C(object):
                    def new_func(self):
                        pass
                def f(arg):
                    arg.new_func()
            """),
            mod1.read(),
        )

    def _true(self, *args):
        return True

    def test_renaming_when_unsure_with_confirmation(self):
        def confirm(occurrence):
            return False

        code = dedent("""\
            class C(object):
                def a_func(self):
                    pass
            def f(arg):
                arg.a_func()
        """)
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(code)
        self._rename(mod1, code.index("a_func"), "new_func", unsure=confirm)
        self.assertEqual(
            dedent("""\
                class C(object):
                    def new_func(self):
                        pass
                def f(arg):
                    arg.a_func()
            """),
            mod1.read(),
        )

    def test_renaming_when_unsure_not_renaming_knowns(self):
        code = dedent("""\
            class C1(object):
                def a_func(self):
                    pass
            class C2(object):
                def a_func(self):
                    pass
            c1 = C1()
            c1.a_func()
            c2 = C2()
            c2.a_func()
        """)
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(code)
        self._rename(mod1, code.index("a_func"), "new_func", unsure=self._true)
        self.assertEqual(
            dedent("""\
                class C1(object):
                    def new_func(self):
                        pass
                class C2(object):
                    def a_func(self):
                        pass
                c1 = C1()
                c1.new_func()
                c2 = C2()
                c2.a_func()
            """),
            mod1.read(),
        )

    def test_renaming_in_strings_and_comments(self):
        code = dedent("""\
            a_var = 1
            # a_var
        """)
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(code)
        self._rename(mod1, code.index("a_var"), "new_var", docs=True)
        self.assertEqual(
            dedent("""\
                new_var = 1
                # new_var
            """),
            mod1.read(),
        )

    def test_not_renaming_in_strings_and_comments_where_not_visible(self):
        code = dedent("""\
            def f():
                a_var = 1
            # a_var
        """)
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(code)
        self._rename(mod1, code.index("a_var"), "new_var", docs=True)
        self.assertEqual(
            dedent("""\
                def f():
                    new_var = 1
                # a_var
            """),
            mod1.read(),
        )

    def test_not_renaming_all_text_occurrences_in_strings_and_comments(self):
        code = dedent("""\
            a_var = 1
            # a_vard _a_var
        """)
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(code)
        self._rename(mod1, code.index("a_var"), "new_var", docs=True)
        self.assertEqual(
            dedent("""\
                new_var = 1
                # a_vard _a_var
            """),
            mod1.read(),
        )

    def test_renaming_occurrences_in_overwritten_scopes(self):
        refactored = self._local_rename(
            dedent("""\
                a_var = 20
                def f():
                    print(a_var)
                def f():
                    print(a_var)
            """),
            2,
            "new_var",
        )
        self.assertEqual(
            dedent("""\
                new_var = 20
                def f():
                    print(new_var)
                def f():
                    print(new_var)
            """),
            refactored,
        )

    def test_renaming_occurrences_in_overwritten_scopes2(self):
        code = dedent("""\
            def f():
                a_var = 1
                print(a_var)
            def f():
                a_var = 1
                print(a_var)
        """)
        refactored = self._local_rename(code, code.index("a_var") + 1, "new_var")
        self.assertEqual(code.replace("a_var", "new_var", 2), refactored)

    @testutils.only_for_versions_higher("3.5")
    def test_renaming_in_generalized_dict_unpacking(self):
        code = dedent("""\
            a_var = {**{'stuff': 'can'}, **{'stuff': 'crayon'}}

            if "stuff" in a_var:
                print("ya")
        """)
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(code)
        refactored = self._local_rename(code, code.index("a_var") + 1, "new_var")
        expected = dedent("""\
            new_var = {**{'stuff': 'can'}, **{'stuff': 'crayon'}}

            if "stuff" in new_var:
                print("ya")
        """)
        self.assertEqual(expected, refactored)

    def test_dos_line_ending_and_renaming(self):
        code = "\r\na = 1\r\n\r\nprint(2 + a + 2)\r\n"
        offset = code.replace("\r\n", "\n").rindex("a")
        refactored = self._local_rename(code, offset, "b")
        self.assertEqual(
            "\nb = 1\n\nprint(2 + b + 2)\n", refactored.replace("\r\n", "\n")
        )

    def test_multi_byte_strs_and_renaming(self):
        s = "{LATIN SMALL LETTER I WITH DIAERESIS}" * 4
        code = "# -*- coding: utf-8 -*-\n# " + s + "\na = 1\nprint(2 + a + 2)\n"
        refactored = self._local_rename(code, code.rindex("a"), "b")
        self.assertEqual(
            "# -*- coding: utf-8 -*-\n# " + s + "\nb = 1\nprint(2 + b + 2)\n",
            refactored,
        )

    def test_resources_parameter(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write(dedent("""\
            def f():
                pass
        """))
        mod2.write(dedent("""\
            import mod1
            mod1.f()
        """))
        self._rename(mod1, mod1.read().rindex("f"), "g", resources=[mod1])
        self.assertEqual(
            dedent("""\
                def g():
                    pass
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                import mod1
                mod1.f()
            """),
            mod2.read(),
        )

    def test_resources_parameter_not_changing_defining_module(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write(dedent("""\
            def f():
                pass
        """))
        mod2.write(dedent("""\
            import mod1
            mod1.f()
        """))
        self._rename(mod1, mod1.read().rindex("f"), "g", resources=[mod2])
        self.assertEqual(
            dedent("""\
                def f():
                    pass
            """),
            mod1.read(),
        )
        self.assertEqual(
            dedent("""\
                import mod1
                mod1.g()
            """),
            mod2.read(),
        )

    # XXX: with variables should not leak
    def xxx_test_with_statement_variables_should_not_leak(self):
        code = dedent("""\
            f = 1
            with open("1.txt") as f:
                print(f)
        """)
        if sys.version_info < (2, 6, 0):
            code = "from __future__ import with_statement\n" + code
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(code)
        self._rename(mod1, code.rindex("f"), "file")
        expected = dedent("""\
            f = 1
            with open("1.txt") as file:
                print(file)
        """)
        self.assertEqual(expected, mod1.read())

    def test_rename_in_list_comprehension(self):
        code = dedent("""\
            some_var = 1
            compr = [some_var for some_var in range(10)]
        """)
        offset = code.index("some_var")
        refactored = self._local_rename(code, offset, "new_var")
        expected = dedent("""\
            new_var = 1
            compr = [some_var for some_var in range(10)]
        """)
        self.assertEqual(refactored, expected)

    def test_renaming_modules_aliased_with_dots(self):
        pkg = testutils.create_package(self.project, "json")
        mod1 = testutils.create_module(self.project, "utils", pkg)

        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            import json.utils as stdlib_json_utils
        """))
        self._rename(pkg, None, "new_json")
        self.assertTrue(
            not mod1.exists() and self.project.find_module("new_json.utils") is not None
        )
        self.assertEqual("import new_json.utils as stdlib_json_utils\n", mod2.read())

    def test_renaming_modules_aliased_many_dots(self):
        pkg = testutils.create_package(self.project, "json")
        mod1 = testutils.create_module(self.project, "utils", pkg)

        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            import json.utils.a as stdlib_json_utils
        """))
        self._rename(pkg, None, "new_json")
        self.assertTrue(
            not mod1.exists() and self.project.find_module("new_json.utils") is not None
        )
        self.assertEqual("import new_json.utils.a as stdlib_json_utils\n", mod2.read())


class ChangeOccurrencesTest(unittest.TestCase):
    def setUp(self):
        self.project = testutils.sample_project()
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_case(self):
        self.mod.write(dedent("""\
            a_var = 1
            print(a_var)
        """))
        changer = rename.ChangeOccurrences(
            self.project, self.mod, self.mod.read().index("a_var")
        )
        changer.get_changes("new_var").do()
        self.assertEqual(
            dedent("""\
                new_var = 1
                print(new_var)
            """),
            self.mod.read(),
        )

    def test_only_performing_inside_scopes(self):
        self.mod.write(dedent("""\
            a_var = 1
            new_var = 2
            def f():
                print(a_var)
        """))
        changer = rename.ChangeOccurrences(
            self.project, self.mod, self.mod.read().rindex("a_var")
        )
        changer.get_changes("new_var").do()
        self.assertEqual(
            dedent("""\
                a_var = 1
                new_var = 2
                def f():
                    print(new_var)
            """),
            self.mod.read(),
        )

    def test_only_performing_on_calls(self):
        self.mod.write(dedent("""\
            def f1():
                pass
            def f2():
                pass
            g = f1
            a = f1()
        """))
        changer = rename.ChangeOccurrences(
            self.project, self.mod, self.mod.read().rindex("f1")
        )
        changer.get_changes("f2", only_calls=True).do()
        self.assertEqual(
            dedent("""\
                def f1():
                    pass
                def f2():
                    pass
                g = f1
                a = f2()
            """),
            self.mod.read(),
        )

    def test_only_performing_on_reads(self):
        self.mod.write(dedent("""\
            a = 1
            b = 2
            print(a)
        """))
        changer = rename.ChangeOccurrences(
            self.project, self.mod, self.mod.read().rindex("a")
        )
        changer.get_changes("b", writes=False).do()
        self.assertEqual(
            dedent("""\
                a = 1
                b = 2
                print(b)
            """),
            self.mod.read(),
        )


class ImplicitInterfacesTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project(validate_objectdb=True)
        self.pycore = self.project.pycore
        self.mod1 = testutils.create_module(self.project, "mod1")
        self.mod2 = testutils.create_module(self.project, "mod2")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _rename(self, resource, offset, new_name, **kwds):
        changes = Rename(self.project, resource, offset).get_changes(new_name, **kwds)
        self.project.do(changes)

    def test_performing_rename_on_parameters(self):
        self.mod1.write("def f(arg):\n    arg.run()\n")
        self.mod2.write(dedent("""\
            import mod1


            class A(object):
                def run(self):
                    pass
            class B(object):
                def run(self):
                    pass
            mod1.f(A())
            mod1.f(B())
        """))
        self.pycore.analyze_module(self.mod2)
        self._rename(self.mod1, self.mod1.read().index("run"), "newrun")
        self.assertEqual("def f(arg):\n    arg.newrun()\n", self.mod1.read())
        self.assertEqual(
            dedent("""\
                import mod1


                class A(object):
                    def newrun(self):
                        pass
                class B(object):
                    def newrun(self):
                        pass
                mod1.f(A())
                mod1.f(B())
            """),
            self.mod2.read(),
        )



print("import: 'rope'")
import rope

print("import: 'rope.base'")
import rope.base

print("import: 'rope.base.oi'")
import rope.base.oi

print("import: 'rope.base.oi.type_hinting'")
import rope.base.oi.type_hinting

print("import: 'rope.base.oi.type_hinting.providers'")
import rope.base.oi.type_hinting.providers

print("import: 'rope.base.oi.type_hinting.resolvers'")
import rope.base.oi.type_hinting.resolvers

print("import: 'rope.base.utils'")
import rope.base.utils

print("import: 'rope.contrib'")
import rope.contrib

print("import: 'rope.contrib.autoimport'")
import rope.contrib.autoimport

print("import: 'rope.refactor'")
import rope.refactor

print("import: 'rope.refactor.importutils'")
import rope.refactor.importutils




import collections
import numbers
import re
import warnings
from itertools import chain

from rope.base import ast, codeanalyze, exceptions

COMMA_IN_WITH_PATTERN = re.compile(r"\(.*?\)|(,)")


def get_patched_ast(source, sorted_children=False):
    """Adds ``region`` and ``sorted_children`` fields to nodes

    Adds ``sorted_children`` field only if `sorted_children` is True.

    """
    return patch_ast(ast.parse(source), source, sorted_children)


def patch_ast(node, source, sorted_children=False):
    """Patches the given node

    After calling, each node in `node` will have a new field named
    `region` that is a tuple containing the start and end offsets
    of the code that generated it.

    If `sorted_children` is true, a `sorted_children` field will
    be created for each node, too.  It is a list containing child
    nodes as well as whitespaces and comments that occur between
    them.

    """
    if hasattr(node, "region"):
        return node
    walker = _PatchingASTWalker(source, children=sorted_children)
    walker(node)
    return node


def node_region(patched_ast_node):
    """Get the region of a patched ast node"""
    return patched_ast_node.region


def write_ast(patched_ast_node):
    """Extract source form a patched AST node with `sorted_children` field

    If the node is patched with sorted_children turned off you can use
    `node_region` function for obtaining code using module source code.
    """
    result = []
    for child in patched_ast_node.sorted_children:
        if isinstance(child, ast.AST):
            result.append(write_ast(child))
        else:
            result.append(child)
    return "".join(result)


class MismatchedTokenError(exceptions.RopeError):
    pass


class _PatchingASTWalker:
    def __init__(self, source, children=False):
        self.source = _Source(source)
        self.children = children
        self.lines = codeanalyze.SourceLinesAdapter(source)
        self.children_stack = []

    Number = object()
    String = object()
    with_or_comma_context_manager = object()
    empty_tuple = object()

    def __call__(self, node):
        method = getattr(self, "_" + node.__class__.__name__, None)
        if method is not None:
            return method(node)
        # ???: Unknown node; what should we do here?
        warnings.warn(
            "Unknown node type <%s>; please report!" % node.__class__.__name__,
            RuntimeWarning,
        )
        node.region = (self.source.offset, self.source.offset)
        if self.children:
            node.sorted_children = [child for field, child in ast.iter_fields(node)]

    def _handle(self, node, base_children, eat_parens=False, eat_spaces=False):
        if hasattr(node, "region"):
            # ???: The same node was seen twice; what should we do?
            warnings.warn(
                "Node <%s> has been already patched; please report!"
                % node.__class__.__name__,
                RuntimeWarning,
            )
            return

        base_children = collections.deque(base_children)
        self.children_stack.append(base_children)
        children = collections.deque()
        formats = []
        suspected_start = self.source.offset
        start = suspected_start
        first_token = True
        while base_children:
            child = base_children.popleft()
            if child is None:
                continue
            offset = self.source.offset
            if isinstance(child, ast.AST):
                self(child)
                token_start = child.region[0]
            else:
                if child is self.String:
                    region = self.source.consume_string(
                        end=self._find_next_statement_start()
                    )
                elif child is self.Number:
                    region = self.source.consume_number()
                elif child == self.empty_tuple:
                    region = self.source.consume_empty_tuple()
                elif child == self.with_or_comma_context_manager:
                    region = self.source.consume_with_or_comma_context_manager()
                elif isinstance(node, (ast.JoinedStr, ast.FormattedValue)):
                    region = self.source.consume_joined_string(child)
                else:
                    region = self.source.consume(child)
                child = self.source[region[0] : region[1]]
                token_start = region[0]
            if not first_token:
                if not isinstance(node, ast.JoinedStr):
                    formats.append(self.source[offset:token_start])
                if self.children:
                    children.append(self.source[offset:token_start])
            else:
                first_token = False
                start = token_start
            if self.children:
                children.append(child)
        start = self._handle_parens(children, start, formats)
        if eat_parens:
            start = self._eat_surrounding_parens(children, suspected_start, start)
        if eat_spaces:
            if self.children:
                children.appendleft(self.source[0:start])
            end_spaces = self.source[self.source.offset :]
            self.source.consume(end_spaces)
            if self.children:
                children.append(end_spaces)
            start = 0
        if self.children:
            node.sorted_children = children
        node.region = (start, self.source.offset)
        self.children_stack.pop()

    def _handle_parens(self, children, start, formats):
        """Changes `children` and returns new start"""
        opens, closes = self._count_needed_parens(formats)
        old_end = self.source.offset
        new_end = None
        for i in range(closes):
            new_end = self.source.consume(")")[1]
        if new_end is not None and self.children:
            children.append(self.source[old_end:new_end])
        new_start = start
        for i in range(opens):
            new_start = self.source.rfind_token("(", 0, new_start)
        if new_start != start:
            if self.children:
                children.appendleft(self.source[new_start:start])
            start = new_start
        return start

    def _eat_surrounding_parens(self, children, suspected_start, start):
        index = self.source.rfind_token("(", suspected_start, start)
        if index is not None:
            old_start = start
            old_offset = self.source.offset
            start = index
            if self.children:
                children.appendleft(self.source[start + 1 : old_start])
                children.appendleft("(")
            token_start, token_end = self.source.consume(")")
            if self.children:
                children.append(self.source[old_offset:token_start])
                children.append(")")
        return start

    def _count_needed_parens(self, children):
        start = 0
        opens = 0
        for child in children:
            if not isinstance(child, (str, bytes)):
                continue
            if child == "" or child[0] in "'\"":
                continue
            index = 0
            while index < len(child):
                if child[index] == ")":
                    if opens > 0:
                        opens -= 1
                    else:
                        start += 1
                if child[index] == "(":
                    opens += 1
                if child[index] == "#":
                    try:
                        index = child.index("\n", index)
                    except ValueError:
                        break
                index += 1
        return start, opens

    def _find_next_statement_start(self):
        for children in reversed(self.children_stack):
            for child in children:
                if isinstance(child, ast.stmt):
                    return child.col_offset + self.lines.get_line_start(child.lineno)
        return len(self.source.source)

    def _join(self, iterable, separator):
        iterable = iter(iterable)
        try:
            yield next(iterable)
        except StopIteration:
            return
        for child in iterable:
            yield separator
            yield child

    def _flatten_keywords(self, iterable):
        iterable = ([attr, "=", pattern] for attr, pattern in iterable)
        iterable = self._join(iterable, separator=[","])
        return chain.from_iterable(iterable)

    def _child_nodes(self, nodes, separator):
        return list(self._join(nodes, separator=separator))

    _operators = {
        "And": "and",
        "Or": "or",
        "Add": "+",
        "Sub": "-",
        "Mult": "*",
        "Div": "/",
        "Mod": "%",
        "Pow": "**",
        "MatMult": "@",
        "LShift": "<<",
        "RShift": ">>",
        "BitOr": "|",
        "BitAnd": "&",
        "BitXor": "^",
        "FloorDiv": "//",
        "Invert": "~",
        "Not": "not",
        "UAdd": "+",
        "USub": "-",
        "Eq": "==",
        "NotEq": "!=",
        "Lt": "<",
        "LtE": "<=",
        "Gt": ">",
        "GtE": ">=",
        "Is": "is",
        "IsNot": "is not",
        "In": "in",
        "NotIn": "not in",
    }

    def _get_op(self, node):
        return self._operators[node.__class__.__name__].split(" ")

    def _Attribute(self, node):
        self._handle(node, [node.value, ".", node.attr])

    def _Assert(self, node):
        children = ["assert", node.test]
        if node.msg:
            children.append(",")
            children.append(node.msg)
        self._handle(node, children)

    def _Assign(self, node):
        children = [*self._child_nodes(node.targets, "="), "=", node.value]
        self._handle(node, children)

    def _AugAssign(self, node):
        children = [node.target, *self._get_op(node.op), "=", node.value]
        self._handle(node, children)

    def _AnnAssign(self, node):
        children = [node.target, ":", node.annotation]
        if node.value is not None:
            children.append("=")
            children.append(node.value)
        self._handle(node, children)

    def _BinOp(self, node):
        children = [node.left] + self._get_op(node.op) + [node.right]
        self._handle(node, children)

    def _BoolOp(self, node):
        self._handle(node, self._child_nodes(node.values, self._get_op(node.op)[0]))

    def _Break(self, node):
        self._handle(node, ["break"])

    def _Call(self, node):
        def _arg_sort_key(node):
            if isinstance(node, ast.keyword):
                return (node.value.lineno, node.value.col_offset)
            return (node.lineno, node.col_offset)

        children = [node.func, "("]
        args = sorted([*node.args, *node.keywords], key=_arg_sort_key)
        children.extend(self._child_nodes(args, ","))
        children.append(")")
        self._handle(node, children)

    def _ClassDef(self, node):
        children = []
        for decorator in node.decorator_list:
            children.extend(("@", decorator))
        children.extend(["class", node.name])
        if node.bases:
            children.append("(")
            children.extend(self._child_nodes(node.bases, ","))
            children.append(")")
        children.append(":")
        children.extend(node.body)
        self._handle(node, children)

    def _Compare(self, node):
        children = []
        children.append(node.left)
        for op, expr in zip(node.ops, node.comparators):
            children.extend(self._get_op(op))
            children.append(expr)
        self._handle(node, children)

    def _Delete(self, node):
        self._handle(node, ["del"] + self._child_nodes(node.targets, ","))

    def _Constant(self, node):
        if isinstance(node.value, (str, bytes)):
            self._handle(node, [self.String])
            return

        if any(node.value is v for v in [True, False, None]):
            self._handle(node, [str(node.value)])
            return

        if isinstance(node.value, numbers.Number):
            self._handle(node, [self.Number])
            return

        if node.value is Ellipsis:
            self._handle(node, ["..."])
            return

        assert False

    def _Num(self, node):
        self._handle(node, [self.Number])

    def _Str(self, node):
        self._handle(node, [self.String])

    def _Bytes(self, node):
        self._handle(node, [self.String])

    def _JoinedStr(self, node):
        def start_quote_char():
            possible_quotes = [
                (self.source.source.find(q, start, end), q) for q in QUOTE_CHARS
            ]
            quote_pos, quote_char = min(
                (pos, q) for pos, q in possible_quotes if pos != -1
            )
            return self.source[start : quote_pos + len(quote_char)]

        def end_quote_char():
            possible_quotes = [
                (self.source.source.rfind(q, start, end), q)
                for q in reversed(QUOTE_CHARS)
            ]
            _, quote_pos, quote_char = max(
                (len(q), pos, q) for pos, q in possible_quotes if pos != -1
            )
            return self.source[end - len(quote_char) : end]

        QUOTE_CHARS = ['"""', "'''", '"', "'"]
        offset = self.source.offset
        start, end = self.source.consume_string(
            end=self._find_next_statement_start(),
        )
        self.source.offset = offset

        children = []
        children.append(start_quote_char())
        for part in node.values:
            if isinstance(part, ast.FormattedValue):
                children.append(part)
        children.append(end_quote_char())
        self._handle(node, children)

    def _FormattedValue(self, node):
        children = []
        children.append("{")
        children.append(node.value)
        if node.format_spec:
            children.append(":")
            for val in node.format_spec.values:
                if isinstance(val, ast.FormattedValue):
                    children.append(val.value)
                else:
                    children.append(val.s)
        children.append("}")
        self._handle(node, children)

    def _Continue(self, node):
        self._handle(node, ["continue"])

    def _Dict(self, node):
        children = []
        children.append("{")
        if node.keys:
            for index, (key, value) in enumerate(zip(node.keys, node.values)):
                if key is None:
                    # PEP-448 dict unpacking: {a: b, **unpack}
                    children.extend(["**", value])
                else:
                    children.extend([key, ":", value])
                if index < len(node.keys) - 1:
                    children.append(",")
        children.append("}")
        self._handle(node, children)

    def _Ellipsis(self, node):
        self._handle(node, ["..."])

    def _Expr(self, node):
        self._handle(node, [node.value])

    def _NamedExpr(self, node):
        children = [node.target, ":=", node.value]
        self._handle(node, children)

    def _ExtSlice(self, node):
        children = []
        for index, dim in enumerate(node.dims):
            if index > 0:
                children.append(",")
            children.append(dim)
        self._handle(node, children)

    def _handle_for_loop_node(self, node, is_async):
        children = ["async", "for"] if is_async else ["for"]
        children.extend([node.target, "in", node.iter, ":"])
        children.extend(node.body)
        if node.orelse:
            children.extend(["else", ":"])
            children.extend(node.orelse)
        self._handle(node, children)

    def _For(self, node):
        self._handle_for_loop_node(node, is_async=False)

    def _AsyncFor(self, node):
        self._handle_for_loop_node(node, is_async=True)

    def _ImportFrom(self, node):
        children = ["from"]
        if node.level:
            children.extend("." * node.level)
        if node.module:
            children.extend(node.module.split("."))
        children.append("import")
        children.extend(self._child_nodes(node.names, ","))
        self._handle(node, children)

    def _alias(self, node):
        children = node.name.split(".")
        if node.asname:
            children.extend(["as", node.asname])
        self._handle(node, children)

    def _handle_function_def_node(self, node, is_async):
        children = []
        for decorator in node.decorator_list:
            children.extend(("@", decorator))
        children.extend(["async", "def"] if is_async else ["def"])
        children.append(node.name)
        children.extend(["(", node.args, ")"])
        children.append(":")
        children.extend(node.body)
        self._handle(node, children)

    def _FunctionDef(self, node):
        self._handle_function_def_node(node, is_async=False)

    def _AsyncFunctionDef(self, node):
        self._handle_function_def_node(node, is_async=True)

    def _arguments(self, node):
        children = []
        args = list(node.args)
        defaults = [None] * (len(args) - len(node.defaults)) + list(node.defaults)
        for index, (arg, default) in enumerate(zip(args, defaults)):
            if index > 0:
                children.append(",")
            self._add_args_to_children(children, arg, default)
        if node.vararg is not None:
            if args:
                children.append(",")
            children.extend(["*", node.vararg.arg])
        if node.kwarg is not None:
            if args or node.vararg is not None:
                children.append(",")
            children.extend(["**", node.kwarg.arg])
        self._handle(node, children)

    def _add_args_to_children(self, children, arg, default):
        if isinstance(arg, (list, tuple)):
            self._add_tuple_parameter(children, arg)
        else:
            children.append(arg)
        if default is not None:
            children.append("=")
            children.append(default)

    def _add_tuple_parameter(self, children, arg):
        children.append("(")
        for index, token in enumerate(arg):
            if index > 0:
                children.append(",")
            if isinstance(token, (list, tuple)):
                self._add_tuple_parameter(children, token)
            else:
                children.append(token)
        children.append(")")

    def _GeneratorExp(self, node):
        children = [node.elt, *node.generators]
        self._handle(node, children, eat_parens=True)

    def _comprehension(self, node):
        children = ["for", node.target, "in", node.iter]
        for if_ in node.ifs:
            children.extend(["if", if_])
        self._handle(node, children)

    def _Global(self, node):
        children = ["global", *self._child_nodes(node.names, ",")]
        self._handle(node, children)

    def _If(self, node):
        children = ["elif"] if self._is_elif(node) else ["if"]
        children.extend([node.test, ":"])
        children.extend(node.body)
        if node.orelse:
            if len(node.orelse) == 1 and self._is_elif(node.orelse[0]):
                pass
            else:
                children.extend(["else", ":"])
            children.extend(node.orelse)
        self._handle(node, children)

    def _is_elif(self, node):
        if not isinstance(node, ast.If):
            return False
        offset = self.lines.get_line_start(node.lineno) + node.col_offset
        word = self.source[offset : offset + 4]
        # XXX: This is a bug; the offset does not point to the first
        alt_word = self.source[offset - 5 : offset - 1]
        return "elif" in (word, alt_word)

    def _IfExp(self, node):
        return self._handle(node, [node.body, "if", node.test, "else", node.orelse])

    def _Import(self, node):
        children = ["import", *self._child_nodes(node.names, ",")]
        self._handle(node, children)

    def _keyword(self, node):
        if node.arg is None:
            children = [node.value]
        else:
            children = [node.arg, "=", node.value]
        self._handle(node, children)

    def _Lambda(self, node):
        self._handle(node, ["lambda", node.args, ":", node.body])

    def _List(self, node):
        self._handle(node, ["[", *self._child_nodes(node.elts, ","), "]"])

    def _ListComp(self, node):
        children = ["[", node.elt, *node.generators, "]"]
        self._handle(node, children)

    def _Set(self, node):
        if node.elts:
            self._handle(node, ["{", *self._child_nodes(node.elts, ","), "}"])
            return
        # Python doesn't have empty set literals
        warnings.warn(
            "Tried to handle empty <Set> literal; please report!", RuntimeWarning
        )
        self._handle(node, ["set(", ")"])

    def _SetComp(self, node):
        children = ["{", node.elt, *node.generators, "}"]
        self._handle(node, children)

    def _DictComp(self, node):
        children = ["{", *[node.key, ":", node.value], *node.generators, "}"]
        self._handle(node, children)

    def _Module(self, node):
        self._handle(node, list(node.body), eat_spaces=True)

    def _Name(self, node):
        self._handle(node, [node.id])

    def _NameConstant(self, node):
        self._handle(node, [str(node.value)])

    def _arg(self, node):
        self._handle(node, [node.arg])

    def _Pass(self, node):
        self._handle(node, ["pass"])

    def _Raise(self, node):
        children = ["raise"]
        if node.exc:
            children.append(node.exc)
        if node.cause:
            children.append("from")
            children.append(node.cause)
        self._handle(node, children)

    def _Return(self, node):
        children = ["return"]
        if node.value:
            children.append(node.value)
        self._handle(node, children)

    def _Index(self, node):
        self._handle(node, [node.value])

    def _Subscript(self, node):
        self._handle(node, [node.value, "[", node.slice, "]"])

    def _Slice(self, node):
        children = []
        if node.lower:
            children.append(node.lower)
        children.append(":")
        if node.upper:
            children.append(node.upper)
        if node.step:
            children.append(":")
            children.append(node.step)
        self._handle(node, children)

    def _TryFinally(self, node):
        # @todo fixme
        is_there_except_handler = False
        not_empty_body = True
        if len(node.finalbody) == 1:
            try:
                is_there_except_handler = isinstance(
                    node.handlers[0], ast.ExceptHandler
                )
                not_empty_body = True
            except IndexError:
                pass
        children = []
        if not_empty_body or not is_there_except_handler:
            children.extend(["try", ":"])
        children.extend(node.body)
        children.extend(node.handlers)
        children.extend(["finally", ":"])
        children.extend(node.finalbody)
        self._handle(node, children)

    def _TryExcept(self, node):
        children = ["try", ":"]
        children.extend(node.body)
        children.extend(node.handlers)
        if node.orelse:
            children.extend(["else", ":"])
            children.extend(node.orelse)
        self._handle(node, children)

    def _Try(self, node):
        if len(node.finalbody):
            self._TryFinally(node)
        else:
            self._TryExcept(node)

    def _ExceptHandler(self, node):
        self._excepthandler(node)

    def _excepthandler(self, node):
        children = ["except"]
        if node.type:
            children.append(node.type)
        if node.name:
            children.append("as")
            children.append(node.name)
        children.append(":")
        children.extend(node.body)

        self._handle(node, children)

    def _Tuple(self, node):
        if node.elts:
            self._handle(node, self._child_nodes(node.elts, ","), eat_parens=True)
        else:
            self._handle(node, [self.empty_tuple])

    def _UnaryOp(self, node):
        children = self._get_op(node.op)
        children.append(node.operand)
        self._handle(node, children)

    def _Await(self, node):
        children = ["await"]
        if node.value:
            children.append(node.value)
        self._handle(node, children)

    def _Yield(self, node):
        children = ["yield"]
        if node.value:
            children.append(node.value)
        self._handle(node, children)

    def _YieldFrom(self, node):
        children = ["yield", "from", node.value]
        self._handle(node, children)

    def _While(self, node):
        children = ["while", node.test, ":"]
        children.extend(node.body)
        if node.orelse:
            children.extend(["else", ":"])
            children.extend(node.orelse)
        self._handle(node, children)

    def _handle_with_node(self, node, is_async):
        children = []

        if is_async:
            children.extend(["async"])
        for item in node.items:
            children.extend([self.with_or_comma_context_manager, item.context_expr])
            if item.optional_vars:
                children.extend(["as", item.optional_vars])
        children.append(":")
        children.extend(node.body)
        self._handle(node, children)

    def _With(self, node):
        self._handle_with_node(node, is_async=False)

    def _AsyncWith(self, node):
        self._handle_with_node(node, is_async=True)

    def _Starred(self, node):
        self._handle(node, [node.value])

    def _Match(self, node):
        children = ["match", node.subject, ":"]
        children.extend(node.cases)
        self._handle(node, children)

    def _match_case(self, node):
        children = ["case", node.pattern]
        if node.guard:
            children.extend(["if", node.guard])
        children.append(":")
        children.extend(node.body)
        self._handle(node, children)

    def _MatchAs(self, node):
        if node.pattern:
            children = [node.pattern, "as", node.name]
        elif node.name is None:
            children = ["_"]
        else:
            children = [node.name]
        self._handle(node, children)

    def _MatchClass(self, node):
        children = [
            node.cls,
            "(",
            *self._child_nodes(node.patterns, ","),
            *self._flatten_keywords(zip(node.kwd_attrs, node.kwd_patterns)),
            ")",
        ]
        self._handle(node, children)

    def _MatchValue(self, node):
        self._handle(node, [node.value])

    def _MatchMapping(self, node):
        children = []
        children.append("{")
        for index, (key, value) in enumerate(zip(node.keys, node.patterns)):
            children.extend([key, ":", value])
            if index < len(node.keys) - 1:
                children.append(",")
        children.append("}")
        self._handle(node, children)


class _Source:
    def __init__(self, source):
        self.source = source
        self.offset = 0

    def consume(self, token, skip_comment=True):
        try:
            while True:
                new_offset = self.source.index(token, self.offset)
                if self._good_token(token, new_offset) or not skip_comment:
                    break
                else:
                    self._skip_comment()
        except (ValueError, TypeError):
            raise MismatchedTokenError(
                f"Token <{token}> at {self._get_location()} cannot be matched"
            )
        self.offset = new_offset + len(token)
        return (new_offset, self.offset)

    def consume_joined_string(self, token):
        new_offset = self.source.index(token, self.offset)
        self.offset = new_offset + len(token)
        return (new_offset, self.offset)

    def consume_string(self, end=None):
        if _Source._string_pattern is None:
            string_pattern = codeanalyze.get_string_pattern()
            formatted_string_pattern = codeanalyze.get_formatted_string_pattern()
            original = r"(?:{})|(?:{})".format(
                string_pattern,
                formatted_string_pattern,
            )
            pattern = r"({})((\s|\\\n|#[^\n]*\n)*({}))*".format(
                original,
                original,
            )
            _Source._string_pattern = re.compile(pattern)
        repattern = _Source._string_pattern
        return self._consume_pattern(repattern, end)

    def consume_number(self):
        if _Source._number_pattern is None:
            _Source._number_pattern = re.compile(self._get_number_pattern())
        repattern = _Source._number_pattern
        return self._consume_pattern(repattern)

    def consume_empty_tuple(self):
        return self._consume_pattern(re.compile(r"\(\s*\)"))

    def consume_with_or_comma_context_manager(self):
        repattern = re.compile(r"with|,")
        return self._consume_pattern(repattern)

    def _good_token(self, token, offset, start=None):
        """Checks whether consumed token is in comments"""
        if start is None:
            start = self.offset
        try:
            comment_index = self.source.rindex("#", start, offset)
        except ValueError:
            return True
        try:
            new_line_index = self.source.rindex("\n", start, offset)
        except ValueError:
            return False
        return comment_index < new_line_index

    def _skip_comment(self):
        self.offset = self.source.index("\n", self.offset + 1)

    def _get_location(self):
        lines = self.source[: self.offset].split("\n")
        return (len(lines), len(lines[-1]))

    def _consume_pattern(self, repattern, end=None):
        while True:
            if end is None:
                end = len(self.source)
            match = repattern.search(self.source, self.offset, end)
            if self._good_token(match.group(), match.start()):
                break
            else:
                self._skip_comment()
        self.offset = match.end()
        return match.start(), match.end()

    def till_token(self, token):
        new_offset = self.source.index(token, self.offset)
        return self[self.offset : new_offset]

    def rfind_token(self, token, start, end):
        index = start
        while True:
            try:
                index = self.source.rindex(token, start, end)
                if self._good_token(token, index, start=start):
                    return index
                else:
                    end = index
            except ValueError:
                return None

    def from_offset(self, offset):
        return self[offset : self.offset]

    def find_backwards(self, pattern, offset):
        return self.source.rindex(pattern, 0, offset)

    def __getitem__(self, index):
        return self.source[index]

    def __getslice__(self, i, j):
        return self.source[i:j]

    def _get_number_pattern(self):
        # HACK: It is merely an approaximation and does the job
        integer = r"\-?(0[xo][\da-fA-F]+|\d+)"
        return r"(%s(\.\d*)?|(\.\d+))([eE][-+]?\d+)?[jJ]?" % integer

    _string_pattern = None
    _number_pattern = None



import sys
import unittest
from textwrap import dedent

from rope.base import exceptions, libutils
from rope.base.builtins import BuiltinClass, File
from rope.base.pycore import _TextChangeDetector
from rope.base.pynamesdef import AssignedName
from rope.base.pyobjects import AbstractFunction, get_base_type
from ropetest import testutils


class PyCoreTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_module(self):
        testutils.create_module(self.project, "mod")
        result = self.project.get_module("mod")
        self.assertEqual(get_base_type("Module"), result.type)
        self.assertEqual(0, len(result.get_attributes()))

    def test_nested_modules(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod = testutils.create_module(self.project, "mod", pkg)  # noqa
        package = self.project.get_module("pkg")
        self.assertEqual(get_base_type("Module"), package.get_type())
        self.assertEqual(1, len(package.get_attributes()))
        module = package["mod"].get_object()
        self.assertEqual(get_base_type("Module"), module.get_type())

    def test_package(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod = testutils.create_module(self.project, "mod", pkg)  # noqa
        result = self.project.get_module("pkg")
        self.assertEqual(get_base_type("Module"), result.type)

    def test_simple_class(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write(dedent("""\
            class SampleClass(object):
                pass
        """))
        mod_element = self.project.get_module("mod")
        result = mod_element["SampleClass"].get_object()
        self.assertEqual(get_base_type("Type"), result.get_type())

    def test_simple_function(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write(dedent("""\
            def sample_function():
                pass
        """))
        mod_element = self.project.get_module("mod")
        result = mod_element["sample_function"].get_object()
        self.assertEqual(get_base_type("Function"), result.get_type())

    def test_class_methods(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class SampleClass(object):
                def sample_method(self):
                    pass
        """)
        mod.write(code)
        mod_element = self.project.get_module("mod")
        sample_class = mod_element["SampleClass"].get_object()
        self.assertTrue("sample_method" in sample_class)
        method = sample_class["sample_method"].get_object()
        self.assertEqual(get_base_type("Function"), method.get_type())

    def test_global_variable_without_type_annotation(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write("var = 10")
        mod_element = self.project.get_module("mod")
        var = mod_element["var"]
        self.assertEqual(AssignedName, type(var))

    @testutils.only_for_versions_higher("3.6")
    def test_global_variable_with_type_annotation(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write("py3_var: str = foo_bar")
        mod_element = self.project.get_module("mod")
        py3_var = mod_element["py3_var"]
        self.assertEqual(AssignedName, type(py3_var))

    def test_class_variables(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write(dedent("""\
            class SampleClass(object):
                var = 10
        """))
        mod_element = self.project.get_module("mod")
        sample_class = mod_element["SampleClass"].get_object()
        var = sample_class["var"]  # noqa

    def test_class_attributes_set_in_init(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write(dedent("""\
            class C(object):
                def __init__(self):
                    self.var = 20
        """))
        mod_element = self.project.get_module("mod")
        sample_class = mod_element["C"].get_object()
        var = sample_class["var"]  # noqa

    def test_class_attributes_set_in_init_overwriting_a_defined(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                def __init__(self):
                    self.f = 20
                def f():
                    pass
        """)
        mod.write(code)
        mod_element = self.project.get_module("mod")
        sample_class = mod_element["C"].get_object()
        f = sample_class["f"].get_object()
        self.assertTrue(isinstance(f, AbstractFunction))

    def test_classes_inside_other_classes(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class SampleClass(object):
                class InnerClass(object):
                    pass

        """)
        mod.write(code)
        mod_element = self.project.get_module("mod")
        sample_class = mod_element["SampleClass"].get_object()
        var = sample_class["InnerClass"].get_object()
        self.assertEqual(get_base_type("Type"), var.get_type())

    def test_non_existent_module(self):
        with self.assertRaises(exceptions.ModuleNotFoundError):
            self.project.get_module("doesnotexistmodule")

    def test_imported_names(self):
        testutils.create_module(self.project, "mod1")
        mod = testutils.create_module(self.project, "mod2")
        mod.write("import mod1\n")
        module = self.project.get_module("mod2")
        imported_sys = module["mod1"].get_object()
        self.assertEqual(get_base_type("Module"), imported_sys.get_type())

    def test_imported_as_names(self):
        testutils.create_module(self.project, "mod1")
        mod = testutils.create_module(self.project, "mod2")
        mod.write("import mod1 as my_import\n")
        module = self.project.get_module("mod2")
        imported_mod = module["my_import"].get_object()
        self.assertEqual(get_base_type("Module"), imported_mod.get_type())

    def test_get_string_module(self):
        code = dedent("""\
            class Sample(object):
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        sample_class = mod["Sample"].get_object()
        self.assertEqual(get_base_type("Type"), sample_class.get_type())

    def test_get_string_module_with_extra_spaces(self):
        code = "a = 10\n    "
        mod = libutils.get_string_module(self.project, code)  # noqa

    def test_parameter_info_for_functions(self):
        code = dedent("""\
            def func(param1, param2=10, *param3, **param4):
                pass""")
        mod = libutils.get_string_module(self.project, code)
        sample_function = mod["func"]
        self.assertEqual(
            ["param1", "param2", "param3", "param4"],
            sample_function.get_object().get_param_names(),
        )

    # FIXME: Not found modules
    def xxx_test_not_found_module_is_module(self):
        code = "import doesnotexist\n"
        mod = libutils.get_string_module(self.project, code)
        self.assertEqual(
            get_base_type("Module"), mod["doesnotexist"].get_object().get_type()
        )

    def test_mixing_scopes_and_objects_hierarchy(self):
        code = "var = 200\n"
        mod = libutils.get_string_module(self.project, code)
        scope = mod.get_scope()
        self.assertTrue("var" in scope.get_names())

    def test_inheriting_base_class_attributes(self):
        code = dedent("""\
            class Base(object):
                def method(self):
                    pass
            class Derived(Base):
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        derived = mod["Derived"].get_object()
        self.assertTrue("method" in derived)
        self.assertEqual(
            get_base_type("Function"), derived["method"].get_object().get_type()
        )

    def test_inheriting_multiple_base_class_attributes(self):
        code = dedent("""\
            class Base1(object):
                def method1(self):
                    pass
            class Base2(object):
                def method2(self):
                    pass
            class Derived(Base1, Base2):
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        derived = mod["Derived"].get_object()
        self.assertTrue("method1" in derived)
        self.assertTrue("method2" in derived)

    def test_inherit_multiple_base_class_attrs_with_the_same_name(self):
        code = dedent("""\
            class Base1(object):
                def method(self):
                    pass
            class Base2(object):
                def method(self):
                    pass
            class Derived(Base1, Base2):
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        base1 = mod["Base1"].get_object()
        derived = mod["Derived"].get_object()
        self.assertEqual(base1["method"].get_object(), derived["method"].get_object())

    def test_inheriting_unknown_base_class(self):
        code = dedent("""\
            class Derived(NotFound):
                def f(self):
                    pass
        """)
        mod = libutils.get_string_module(self.project, code)
        derived = mod["Derived"].get_object()
        self.assertTrue("f" in derived)

    def test_module_creation(self):
        new_module = testutils.create_module(self.project, "module")
        self.assertFalse(new_module.is_folder())
        self.assertEqual(self.project.get_resource("module.py"), new_module)

    def test_packaged_module_creation(self):
        package = self.project.root.create_folder("package")  # noqa
        new_module = testutils.create_module(self.project, "package.module")
        self.assertEqual(self.project.get_resource("package/module.py"), new_module)

    def test_packaged_module_creation_with_nested_src(self):
        src = self.project.root.create_folder("src")
        src.create_folder("pkg")
        new_module = testutils.create_module(self.project, "pkg.mod", src)
        self.assertEqual(self.project.get_resource("src/pkg/mod.py"), new_module)

    def test_package_creation(self):
        new_package = testutils.create_package(self.project, "pkg")
        self.assertTrue(new_package.is_folder())
        self.assertEqual(self.project.get_resource("pkg"), new_package)
        self.assertEqual(
            self.project.get_resource("pkg/__init__.py"),
            new_package.get_child("__init__.py"),
        )

    def test_nested_package_creation(self):
        testutils.create_package(self.project, "pkg1")
        nested_package = testutils.create_package(self.project, "pkg1.pkg2")
        self.assertEqual(self.project.get_resource("pkg1/pkg2"), nested_package)

    def test_packaged_package_creation_with_nested_src(self):
        src = self.project.root.create_folder("src")
        testutils.create_package(self.project, "pkg1", src)
        nested_package = testutils.create_package(self.project, "pkg1.pkg2", src)
        self.assertEqual(self.project.get_resource("src/pkg1/pkg2"), nested_package)

    def test_find_module(self):
        src = self.project.root.create_folder("src")
        samplemod = testutils.create_module(self.project, "samplemod", src)
        found_module = self.project.find_module("samplemod")
        self.assertEqual(samplemod, found_module)

    def test_find_nested_module(self):
        src = self.project.root.create_folder("src")
        samplepkg = testutils.create_package(self.project, "samplepkg", src)
        samplemod = testutils.create_module(self.project, "samplemod", samplepkg)
        found_module = self.project.find_module("samplepkg.samplemod")
        self.assertEqual(samplemod, found_module)

    def test_find_multiple_module(self):
        src = self.project.root.create_folder("src")
        samplemod1 = testutils.create_module(self.project, "samplemod", src)
        samplemod2 = testutils.create_module(self.project, "samplemod")
        test = self.project.root.create_folder("test")
        samplemod3 = testutils.create_module(self.project, "samplemod", test)
        found_module = self.project.find_module("samplemod")
        self.assertTrue(
            samplemod1 == found_module
            or samplemod2 == found_module
            or samplemod3 == found_module
        )

    def test_find_module_packages(self):
        src = self.project.root
        samplepkg = testutils.create_package(self.project, "samplepkg", src)
        found_module = self.project.find_module("samplepkg")
        self.assertEqual(samplepkg, found_module)

    def test_find_module_when_module_and_package_with_the_same_name(self):
        src = self.project.root
        testutils.create_module(self.project, "sample", src)
        samplepkg = testutils.create_package(self.project, "sample", src)
        found_module = self.project.find_module("sample")
        self.assertEqual(samplepkg, found_module)

    def test_source_folders_preference(self):
        testutils.create_package(self.project, "pkg1")
        testutils.create_package(self.project, "pkg1.src2")
        lost = testutils.create_module(self.project, "pkg1.src2.lost")
        self.assertEqual(self.project.find_module("lost"), None)
        self.project.close()
        from rope.base.project import Project

        self.project = Project(self.project.address, source_folders=["pkg1/src2"])
        self.assertEqual(self.project.find_module("lost"), lost)

    def test_get_pyname_definition_location(self):
        code = "a_var = 20\n"
        mod = libutils.get_string_module(self.project, code)
        a_var = mod["a_var"]
        self.assertEqual((mod, 1), a_var.get_definition_location())

    def test_get_pyname_definition_location_functions(self):
        code = dedent("""\
            def a_func():
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        a_func = mod["a_func"]
        self.assertEqual((mod, 1), a_func.get_definition_location())

    def test_get_pyname_definition_location_class(self):
        code = dedent("""\
            class AClass(object):
                pass

        """)
        mod = libutils.get_string_module(self.project, code)
        a_class = mod["AClass"]
        self.assertEqual((mod, 1), a_class.get_definition_location())

    def test_get_pyname_definition_location_local_variables(self):
        code = dedent("""\
            def a_func():
                a_var = 10
        """)
        mod = libutils.get_string_module(self.project, code)
        a_func_scope = mod.get_scope().get_scopes()[0]
        a_var = a_func_scope["a_var"]
        self.assertEqual((mod, 2), a_var.get_definition_location())

    def test_get_pyname_definition_location_reassigning(self):
        code = dedent("""\
            a_var = 20
            a_var=30
        """)
        mod = libutils.get_string_module(self.project, code)
        a_var = mod["a_var"]
        self.assertEqual((mod, 1), a_var.get_definition_location())

    def test_get_pyname_definition_location_imported(self):
        testutils.create_module(self.project, "mod")
        code = "import mod\n"
        mod = libutils.get_string_module(self.project, code)
        imported_module = self.project.get_module("mod")
        module_pyname = mod["mod"]
        self.assertEqual((imported_module, 1), module_pyname.get_definition_location())

    def test_get_pyname_definition_location_imports(self):
        module_resource = testutils.create_module(self.project, "mod")
        module_resource.write(dedent("""\

            def a_func():
                pass
        """))
        imported_module = self.project.get_module("mod")
        code = dedent("""\
            from mod import a_func
        """)
        mod = libutils.get_string_module(self.project, code)
        a_func = mod["a_func"]
        self.assertEqual((imported_module, 2), a_func.get_definition_location())

    def test_get_pyname_definition_location_parameters(self):
        code = dedent("""\
            def a_func(param1, param2):
                a_var = param
        """)
        mod = libutils.get_string_module(self.project, code)
        a_func_scope = mod.get_scope().get_scopes()[0]
        param1 = a_func_scope["param1"]
        self.assertEqual((mod, 1), param1.get_definition_location())
        param2 = a_func_scope["param2"]
        self.assertEqual((mod, 1), param2.get_definition_location())

    def test_module_get_resource(self):
        module_resource = testutils.create_module(self.project, "mod")
        module = self.project.get_module("mod")
        self.assertEqual(module_resource, module.get_resource())
        code = dedent("""\
            from mod import a_func
        """)
        string_module = libutils.get_string_module(self.project, code)
        self.assertEqual(None, string_module.get_resource())

    def test_get_pyname_definition_location_class2(self):
        code = dedent("""\
            class AClass(object):
                def __init__(self):
                    self.an_attr = 10
        """)
        mod = libutils.get_string_module(self.project, code)
        a_class = mod["AClass"].get_object()
        an_attr = a_class["an_attr"]
        self.assertEqual((mod, 3), an_attr.get_definition_location())

    def test_import_not_found_module_get_definition_location(self):
        code = "import doesnotexist\n"
        mod = libutils.get_string_module(self.project, code)
        does_not_exist = mod["doesnotexist"]
        self.assertEqual((None, None), does_not_exist.get_definition_location())

    def test_from_not_found_module_get_definition_location(self):
        code = "from doesnotexist import Sample\n"
        mod = libutils.get_string_module(self.project, code)
        sample = mod["Sample"]
        self.assertEqual((None, None), sample.get_definition_location())

    def test_from_package_import_module_get_definition_location(self):
        pkg = testutils.create_package(self.project, "pkg")
        testutils.create_module(self.project, "mod", pkg)
        pkg_mod = self.project.get_module("pkg.mod")
        code = "from pkg import mod\n"
        mod = libutils.get_string_module(self.project, code)
        imported_mod = mod["mod"]
        self.assertEqual((pkg_mod, 1), imported_mod.get_definition_location())

    def test_get_module_for_defined_pyobjects(self):
        code = dedent("""\
            class AClass(object):
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        a_class = mod["AClass"].get_object()
        self.assertEqual(mod, a_class.get_module())

    def test_get_definition_location_for_packages(self):
        testutils.create_package(self.project, "pkg")
        init_module = self.project.get_module("pkg.__init__")
        code = "import pkg\n"
        mod = libutils.get_string_module(self.project, code)
        pkg_pyname = mod["pkg"]
        self.assertEqual((init_module, 1), pkg_pyname.get_definition_location())

    def test_get_definition_location_for_filtered_packages(self):
        pkg = testutils.create_package(self.project, "pkg")
        testutils.create_module(self.project, "mod", pkg)
        init_module = self.project.get_module("pkg.__init__")
        code = "import pkg.mod"
        mod = libutils.get_string_module(self.project, code)
        pkg_pyname = mod["pkg"]
        self.assertEqual((init_module, 1), pkg_pyname.get_definition_location())

    def test_out_of_project_modules(self):
        code = "import rope.base.project as project\n"
        scope = libutils.get_string_scope(self.project, code)
        imported_module = scope["project"].get_object()
        self.assertTrue("Project" in imported_module)

    def test_file_encoding_reading(self):
        contents = (
            "# -*- coding: utf-8 -*-\n" + "#\N{LATIN SMALL LETTER I WITH DIAERESIS}\n"
        )
        mod = testutils.create_module(self.project, "mod")
        mod.write(contents)
        self.project.get_module("mod")

    def test_global_keyword(self):
        code = dedent("""\
            a_var = 1
            def a_func():
                global a_var
        """)
        mod = libutils.get_string_module(self.project, code)
        global_var = mod["a_var"]
        func_scope = mod["a_func"].get_object().get_scope()
        local_var = func_scope["a_var"]
        self.assertEqual(global_var, local_var)

    def test_not_leaking_for_vars_inside_parent_scope(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                def f(self):
                    for my_var1, my_var2 in []:
                        pass
        """)
        mod.write(code)
        pymod = self.pycore.resource_to_pyobject(mod)
        c_class = pymod["C"].get_object()
        self.assertFalse("my_var1" in c_class)
        self.assertFalse("my_var2" in c_class)

    def test_not_leaking_for_vars_inside_parent_scope2(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                def f(self):
                    for my_var in []:
                        pass
        """)
        mod.write(code)
        pymod = self.pycore.resource_to_pyobject(mod)
        c_class = pymod["C"].get_object()
        self.assertFalse("my_var" in c_class)

    def test_variables_defined_in_excepts(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            try:
                myvar1 = 1
            except:
                myvar2 = 1
            finally:
                myvar3 = 1
        """)
        mod.write(code)
        pymod = self.pycore.resource_to_pyobject(mod)
        self.assertTrue("myvar1" in pymod)
        self.assertTrue("myvar2" in pymod)
        self.assertTrue("myvar3" in pymod)

    def test_not_leaking_tuple_assigned_names_inside_parent_scope(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                def f(self):
                    var1, var2 = range(2)
        """)
        mod.write(code)
        pymod = self.pycore.resource_to_pyobject(mod)
        c_class = pymod["C"].get_object()
        self.assertFalse("var1" in c_class)

    def test_with_statement_variables(self):
        code = dedent("""\
            import threading
            with threading.lock() as var:    pass
        """)
        if sys.version_info < (2, 6, 0):
            code = "from __future__ import with_statement\n" + code
        pymod = libutils.get_string_module(self.project, code)
        self.assertTrue("var" in pymod)

    def test_with_statement_variables_and_tuple_assignment(self):
        code = dedent("""\
            class A(object):
                def __enter__(self):        return (1, 2)
                def __exit__(self, type, value, tb):
                    pass
            with A() as (a, b):
                pass
        """)
        if sys.version_info < (2, 6, 0):
            code = "from __future__ import with_statement\n" + code
        pymod = libutils.get_string_module(self.project, code)
        self.assertTrue("a" in pymod)
        self.assertTrue("b" in pymod)

    def test_with_statement_variable_type(self):
        code = dedent("""\
            class A(object):
                def __enter__(self):
                    return self
                def __exit__(self, type, value, tb):
                    pass
            with A() as var:
                pass
        """)
        if sys.version_info < (2, 6, 0):
            code = "from __future__ import with_statement\n" + code
        pymod = libutils.get_string_module(self.project, code)
        a_class = pymod["A"].get_object()
        var = pymod["var"].get_object()
        self.assertEqual(a_class, var.get_type())

    def test_nested_with_statement_variable_type(self):
        code = dedent("""\
            class A(object):
                def __enter__(self):
                    return self
                def __exit__(self, type, value, tb):
                    pass
            class B(object):
                def __enter__(self):
                    return self
                def __exit__(self, type, value, tb):
                    pass
            with A() as var_a, B() as var_b:
                pass
        """)
        if sys.version_info < (2, 6, 0):
            code = "from __future__ import with_statement\n" + code
        pymod = libutils.get_string_module(self.project, code)
        a_class = pymod["A"].get_object()
        var_a = pymod["var_a"].get_object()
        self.assertEqual(a_class, var_a.get_type())

        b_class = pymod["B"].get_object()
        var_b = pymod["var_b"].get_object()
        self.assertEqual(b_class, var_b.get_type())

    def test_with_statement_with_no_vars(self):
        code = dedent("""\
            with open("file"):    pass
        """)
        if sys.version_info < (2, 6, 0):
            code = "from __future__ import with_statement\n" + code
        pymod = libutils.get_string_module(self.project, code)
        pymod.get_attributes()

    def test_with_statement(self):
        code = dedent("""\
            a = 10
            with open("file") as f:    pass
        """)
        pymod = libutils.get_string_module(self.project, code)
        assigned = pymod.get_attribute("a")
        self.assertEqual(BuiltinClass, type(assigned.get_object().get_type()))

        assigned = pymod.get_attribute("f")
        self.assertEqual(File, type(assigned.get_object().get_type()))

    def test_check_for_else_block(self):
        code = dedent("""\
            for i in range(10):
                pass
            else:
                myvar = 1
        """)
        mod = libutils.get_string_module(self.project, code)
        a_var = mod["myvar"]
        self.assertEqual((mod, 4), a_var.get_definition_location())

    def test_check_names_defined_in_whiles(self):
        code = dedent("""\
            while False:
                myvar = 1
        """)
        mod = libutils.get_string_module(self.project, code)
        a_var = mod["myvar"]
        self.assertEqual((mod, 2), a_var.get_definition_location())

    def test_get_definition_location_in_tuple_assnames(self):
        code = dedent("""\
            def f(x):
                x.z, a = range(2)
        """)
        mod = libutils.get_string_module(self.project, code)
        x = mod["f"].get_object().get_scope()["x"]
        a = mod["f"].get_object().get_scope()["a"]
        self.assertEqual((mod, 1), x.get_definition_location())
        self.assertEqual((mod, 2), a.get_definition_location())

    def test_syntax_errors_in_code(self):
        with self.assertRaises(exceptions.ModuleSyntaxError):
            libutils.get_string_module(self.project, "xyx print\n")

    def test_holding_error_location_information(self):
        try:
            libutils.get_string_module(self.project, "xyx print\n")
        except exceptions.ModuleSyntaxError as e:
            self.assertEqual(1, e.lineno)

    def test_no_exceptions_on_module_encoding_problems(self):
        mod = testutils.create_module(self.project, "mod")
        contents = b"\nsdsdsd\n\xa9\n"
        file = open(mod.real_path, "wb")
        file.write(contents)
        file.close()
        mod.read()

    def test_syntax_errors_when_cannot_decode_file2(self):
        mod = testutils.create_module(self.project, "mod")
        contents = b"\n\xa9\n"
        file = open(mod.real_path, "wb")
        file.write(contents)
        file.close()
        with self.assertRaises(exceptions.ModuleSyntaxError):
            self.pycore.resource_to_pyobject(mod)

    def test_syntax_errors_when_null_bytes(self):
        mod = testutils.create_module(self.project, "mod")
        contents = b"\n\x00\n"
        file = open(mod.real_path, "wb")
        file.write(contents)
        file.close()
        with self.assertRaises(exceptions.ModuleSyntaxError):
            self.pycore.resource_to_pyobject(mod)

    def test_syntax_errors_when_bad_strs(self):
        mod = testutils.create_module(self.project, "mod")
        contents = b'\n"\\x0"\n'
        file = open(mod.real_path, "wb")
        file.write(contents)
        file.close()
        with self.assertRaises(exceptions.ModuleSyntaxError):
            self.pycore.resource_to_pyobject(mod)

    def test_not_reaching_maximum_recursions_with_from_star_imports(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("from mod2 import *\n")
        mod2.write("from mod1 import *\n")
        pymod1 = self.pycore.resource_to_pyobject(mod1)
        pymod1.get_attributes()

    def test_not_reaching_maximum_recursions_when_importing_variables(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("from mod2 import myvar\n")
        mod2.write("from mod1 import myvar\n")
        pymod1 = self.pycore.resource_to_pyobject(mod1)
        pymod1["myvar"].get_object()

    def test_not_reaching_maximum_recursions_when_importing_variables2(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write("from mod1 import myvar\n")
        pymod1 = self.pycore.resource_to_pyobject(mod1)
        pymod1["myvar"].get_object()

    def test_pyobject_equality_should_compare_types(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            var1 = ""
            var2 = ""
        """))
        pymod1 = self.pycore.resource_to_pyobject(mod1)
        self.assertEqual(pymod1["var1"].get_object(), pymod1["var2"].get_object())


class PyCoreInProjectsTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        samplemod = testutils.create_module(self.project, "samplemod")
        code = dedent("""\
            class SampleClass(object):
                def sample_method():
                    pass

            def sample_func():
                pass
            sample_var = 10

            def _underlined_func():
                pass

        """)
        samplemod.write(code)
        package = testutils.create_package(self.project, "package")
        testutils.create_module(self.project, "nestedmod", package)

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_import(self):
        code = "import samplemod\n"
        mod = libutils.get_string_module(self.project, code)
        samplemod = mod["samplemod"].get_object()
        self.assertEqual(get_base_type("Module"), samplemod.get_type())

    def test_from_import_class(self):
        code = "from samplemod import SampleClass\n"
        mod = libutils.get_string_module(self.project, code)
        result = mod["SampleClass"].get_object()
        self.assertEqual(get_base_type("Type"), result.get_type())
        self.assertTrue("sample_func" not in mod.get_attributes())

    def test_from_import_star(self):
        code = "from samplemod import *\n"
        mod = libutils.get_string_module(self.project, code)
        self.assertEqual(
            get_base_type("Type"), mod["SampleClass"].get_object().get_type()
        )
        self.assertEqual(
            get_base_type("Function"), mod["sample_func"].get_object().get_type()
        )
        self.assertTrue(mod["sample_var"] is not None)

    def test_from_import_star_overwriting(self):
        code = dedent("""\
            from samplemod import *
            class SampleClass(object):
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        samplemod = self.project.get_module("samplemod")
        sample_class = samplemod["SampleClass"].get_object()
        self.assertNotEqual(
            sample_class, mod.get_attributes()["SampleClass"].get_object()
        )

    def test_from_import_star_not_importing_underlined(self):
        code = "from samplemod import *"
        mod = libutils.get_string_module(self.project, code)
        self.assertTrue("_underlined_func" not in mod.get_attributes())

    def test_from_import_star_imports_in_functions(self):
        code = dedent("""\
            def f():
                from os import *
        """)
        mod = libutils.get_string_module(self.project, code)
        mod["f"].get_object().get_scope().get_names()

    def test_from_package_import_mod(self):
        code = "from package import nestedmod\n"
        mod = libutils.get_string_module(self.project, code)
        self.assertEqual(
            get_base_type("Module"), mod["nestedmod"].get_object().get_type()
        )

    # XXX: Deciding to import everything on import start from packages
    def xxx_test_from_package_import_star(self):
        code = "from package import *\n"
        mod = libutils.get_string_module(self.project, code)
        self.assertTrue("nestedmod" not in mod.get_attributes())

    def test_unknown_when_module_cannot_be_found(self):
        code = "from doesnotexist import nestedmod\n"
        mod = libutils.get_string_module(self.project, code)
        self.assertTrue("nestedmod" in mod)

    def test_from_import_function(self):
        code = dedent("""\
            def f():
                from samplemod import SampleClass
        """)
        scope = libutils.get_string_scope(self.project, code)
        self.assertEqual(
            get_base_type("Type"),
            scope.get_scopes()[0]["SampleClass"].get_object().get_type(),
        )

    def test_circular_imports(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("import mod2\n")
        mod2.write("import mod1\n")
        self.project.get_module("mod1")

    def test_circular_imports2(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write(dedent("""\
            from mod2 import Sample2
            class Sample1(object):
                pass
        """))
        mod2.write(dedent("""\
            from mod1 import Sample1
            class Sample2(object):
                pass
        """))
        self.project.get_module("mod1").get_attributes()

    def test_multi_dot_imports(self):
        pkg = testutils.create_package(self.project, "pkg")
        pkg_mod = testutils.create_module(self.project, "mod", pkg)
        pkg_mod.write(dedent("""\
            def sample_func():
                pass
        """))
        code = "import pkg.mod\n"
        mod = libutils.get_string_module(self.project, code)
        self.assertTrue("pkg" in mod)
        self.assertTrue("sample_func" in mod["pkg"].get_object()["mod"].get_object())

    def test_multi_dot_imports2(self):
        pkg = testutils.create_package(self.project, "pkg")
        testutils.create_module(self.project, "mod1", pkg)
        testutils.create_module(self.project, "mod2", pkg)
        code = dedent("""\
            import pkg.mod1
            import pkg.mod2
        """)
        mod = libutils.get_string_module(self.project, code)
        package = mod["pkg"].get_object()
        self.assertEqual(2, len(package.get_attributes()))
        self.assertTrue("mod1" in package and "mod2" in package)

    def test_multi_dot_imports3(self):
        pkg1 = testutils.create_package(self.project, "pkg1")
        pkg2 = testutils.create_package(self.project, "pkg2", pkg1)
        testutils.create_module(self.project, "mod1", pkg2)
        testutils.create_module(self.project, "mod2", pkg2)
        code = dedent("""\
            import pkg1.pkg2.mod1
            import pkg1.pkg2.mod2
        """)
        mod = libutils.get_string_module(self.project, code)
        package1 = mod["pkg1"].get_object()
        package2 = package1["pkg2"].get_object()
        self.assertEqual(2, len(package2.get_attributes()))
        self.assertTrue("mod1" in package2 and "mod2" in package2)

    def test_multi_dot_imports_as(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1", pkg)
        mod1.write(dedent("""\
            def f():
                pass
        """))
        code = "import pkg.mod1 as mod1\n"
        mod = libutils.get_string_module(self.project, code)
        module = mod["mod1"].get_object()
        self.assertTrue("f" in module)

    # TODO: not showing unimported names as attributes of packages
    def xxx_test_from_package_import_package(self):
        pkg1 = testutils.create_package(self.project, "pkg1")
        pkg2 = testutils.create_package(self.project, "pkg2", pkg1)
        testutils.create_module(self.project, "mod", pkg2)
        code = "from pkg1 import pkg2\n"
        mod = libutils.get_string_module(self.project, code)
        package = mod["pkg2"]
        self.assertEqual(0, len(package.get_attributes()))

    def test_invalidating_cache_after_resource_change(self):
        module = testutils.create_module(self.project, "mod")
        module.write("import sys\n")
        mod1 = self.project.get_module("mod")
        self.assertTrue("var" not in mod1.get_attributes())
        module.write("var = 10\n")
        mod2 = self.project.get_module("mod")
        self.assertTrue("var" in mod2)

    def test_invalidating_cache_after_resource_change_for_init_dot_pys(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod = testutils.create_module(self.project, "mod")
        init_dot_py = pkg.get_child("__init__.py")
        init_dot_py.write("a_var = 10\n")
        mod.write("import pkg\n")
        pymod = self.project.get_module("mod")
        self.assertTrue("a_var" in pymod["pkg"].get_object())
        init_dot_py.write("new_var = 10\n")
        self.assertTrue("a_var" not in pymod["pkg"].get_object().get_attributes())

    def test_invalidating_cache_after_rsrc_chng_for_nested_init_dot_pys(self):
        pkg1 = testutils.create_package(self.project, "pkg1")
        pkg2 = testutils.create_package(self.project, "pkg2", pkg1)
        mod = testutils.create_module(self.project, "mod")
        init_dot_py = pkg2.get_child("__init__.py")
        init_dot_py.write("a_var = 10\n")
        mod.write("import pkg1\n")
        pymod = self.project.get_module("mod")
        self.assertTrue("a_var" in pymod["pkg1"].get_object()["pkg2"].get_object())
        init_dot_py.write("new_var = 10\n")
        self.assertTrue("a_var" not in pymod["pkg1"].get_object()["pkg2"].get_object())

    def test_from_import_nonexistent_module(self):
        code = "from doesnotexistmod import DoesNotExistClass\n"
        mod = libutils.get_string_module(self.project, code)
        self.assertTrue("DoesNotExistClass" in mod)
        self.assertEqual(
            get_base_type("Unknown"), mod["DoesNotExistClass"].get_object().get_type()
        )

    def test_from_import_nonexistent_name(self):
        code = "from samplemod import DoesNotExistClass\n"
        mod = libutils.get_string_module(self.project, code)
        self.assertTrue("DoesNotExistClass" in mod)
        self.assertEqual(
            get_base_type("Unknown"), mod["DoesNotExistClass"].get_object().get_type()
        )

    def test_not_considering_imported_names_as_sub_scopes(self):
        code = "from samplemod import SampleClass\n"
        scope = libutils.get_string_scope(self.project, code)
        self.assertEqual(0, len(scope.get_scopes()))

    def test_not_considering_imported_modules_as_sub_scopes(self):
        code = "import samplemod\n"
        scope = libutils.get_string_scope(self.project, code)
        self.assertEqual(0, len(scope.get_scopes()))

    def test_inheriting_dotted_base_class(self):
        code = dedent("""\
            import samplemod
            class Derived(samplemod.SampleClass):
                pass
        """)
        mod = libutils.get_string_module(self.project, code)
        derived = mod["Derived"].get_object()
        self.assertTrue("sample_method" in derived)

    def test_self_in_methods(self):
        code = dedent("""\
            class Sample(object):
                def func(self):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)
        sample_class = scope["Sample"].get_object()
        func_scope = scope.get_scopes()[0].get_scopes()[0]
        self.assertEqual(sample_class, func_scope["self"].get_object().get_type())
        self.assertTrue("func" in func_scope["self"].get_object())

    def test_none_assignments_in_classes(self):
        code = dedent("""\
            class C(object):
                var = ""
                def f(self):
                    self.var += "".join([])
        """)
        scope = libutils.get_string_scope(self.project, code)
        c_class = scope["C"].get_object()
        self.assertTrue("var" in c_class)

    def test_self_in_methods_with_decorators(self):
        code = dedent("""\
            class Sample(object):
                @staticmethod
                def func(self):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)
        sample_class = scope["Sample"].get_object()
        func_scope = scope.get_scopes()[0].get_scopes()[0]
        self.assertNotEqual(sample_class, func_scope["self"].get_object().get_type())

    def test_location_of_imports_when_importing(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write("from samplemod import SampleClass\n")
        code = "from mod import SampleClass\n"
        scope = libutils.get_string_scope(self.project, code)
        sample_class = scope["SampleClass"]
        samplemod = self.project.get_module("samplemod")
        self.assertEqual((samplemod, 1), sample_class.get_definition_location())

    def test_nested_modules(self):
        pkg = testutils.create_package(self.project, "pkg")
        testutils.create_module(self.project, "mod", pkg)
        imported_module = self.project.get_module("pkg.mod")
        code = "import pkg.mod\n"
        scope = libutils.get_string_scope(self.project, code)
        mod_pyobject = scope["pkg"].get_object()["mod"]
        self.assertEqual((imported_module, 1), mod_pyobject.get_definition_location())

    def test_reading_init_dot_py(self):
        pkg = testutils.create_package(self.project, "pkg")
        init_dot_py = pkg.get_child("__init__.py")
        init_dot_py.write("a_var = 1\n")
        pkg_object = self.project.get_module("pkg")
        self.assertTrue("a_var" in pkg_object)

    def test_relative_imports(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1", pkg)
        mod2 = testutils.create_module(self.project, "mod2", pkg)
        mod2.write("import mod1\n")
        mod1_object = self.pycore.resource_to_pyobject(mod1)
        mod2_object = self.pycore.resource_to_pyobject(mod2)
        self.assertEqual(mod1_object, mod2_object.get_attributes()["mod1"].get_object())

    def test_relative_froms(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1", pkg)
        mod2 = testutils.create_module(self.project, "mod2", pkg)
        mod1.write(dedent("""\
            def a_func():
                pass
        """))
        mod2.write("from mod1 import a_func\n")
        mod1_object = self.pycore.resource_to_pyobject(mod1)
        mod2_object = self.pycore.resource_to_pyobject(mod2)
        self.assertEqual(
            mod1_object["a_func"].get_object(), mod2_object["a_func"].get_object()
        )

    def test_relative_imports_for_string_modules(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1", pkg)
        mod2 = testutils.create_module(self.project, "mod2", pkg)
        mod2.write("import mod1\n")
        mod1_object = self.pycore.resource_to_pyobject(mod1)
        mod2_object = libutils.get_string_module(self.project, mod2.read(), mod2)
        self.assertEqual(mod1_object, mod2_object["mod1"].get_object())

    def test_relative_imports_for_string_scopes(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1", pkg)
        mod2 = testutils.create_module(self.project, "mod2", pkg)
        mod2.write("import mod1\n")
        mod1_object = self.pycore.resource_to_pyobject(mod1)
        mod2_scope = libutils.get_string_scope(self.project, mod2.read(), mod2)
        self.assertEqual(mod1_object, mod2_scope["mod1"].get_object())

    def test_new_style_relative_imports(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1", pkg)
        mod2 = testutils.create_module(self.project, "mod2", pkg)
        mod2.write("from . import mod1\n")
        mod1_object = self.pycore.resource_to_pyobject(mod1)
        mod2_object = self.pycore.resource_to_pyobject(mod2)
        self.assertEqual(mod1_object, mod2_object["mod1"].get_object())

    def test_new_style_relative_imports2(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2", pkg)
        mod1.write(dedent("""\
            def a_func():
                pass
        """))
        mod2.write("from ..mod1 import a_func\n")
        mod1_object = self.pycore.resource_to_pyobject(mod1)
        mod2_object = self.pycore.resource_to_pyobject(mod2)
        self.assertEqual(
            mod1_object["a_func"].get_object(), mod2_object["a_func"].get_object()
        )

    def test_invalidating_cache_for_from_imports_after_resource_change(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write(dedent("""\
            def a_func():
                print(1)
        """))
        mod1.write(dedent("""\
            from mod2 import a_func
            a_func()
        """))

        pymod1 = self.project.get_module("mod1")
        pymod2 = self.project.get_module("mod2")
        self.assertEqual(pymod1["a_func"].get_object(), pymod2["a_func"].get_object())
        mod2.write(mod2.read() + "\n")
        pymod2 = self.project.get_module("mod2")
        self.assertEqual(pymod1["a_func"].get_object(), pymod2["a_func"].get_object())

    def test_invalidating_superclasses_after_change(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write(dedent("""\
            class A(object):
                def func1(self):
                    pass
        """))
        mod2.write(dedent("""\
            import mod1
            class B(mod1.A):
                pass
        """))

        b_class = self.project.get_module("mod2")["B"].get_object()
        self.assertTrue("func1" in b_class)

        mod1.write(dedent("""\
            class A(object):
                def func2(self):
                    pass
        """))
        self.assertTrue("func2" in b_class)

    def test_caching_pymodule_with_syntax_errors(self):
        self.project.prefs["ignore_syntax_errors"] = True
        self.project.prefs["automatic_soa"] = True
        self.project.pycore._init_automatic_soa()
        source = dedent("""\
            import sys
            ab cd
        """)
        mod = testutils.create_module(self.project, "mod")
        mod.write(source)
        from rope.contrib import fixsyntax

        fixer = fixsyntax.FixSyntax(self.project, source, mod, 10)
        pymodule = fixer.get_pymodule()
        self.assertTrue(pymodule.source_code.startswith("import sys\npass\n"))


class TextChangeDetectorTest(unittest.TestCase):
    def test_trivial_case(self):
        detector = _TextChangeDetector("\n", "\n")
        self.assertFalse(detector.is_changed(1, 1))

    def test_one_line_change(self):
        detector = _TextChangeDetector("1\n2\n", "1\n3\n")
        self.assertFalse(detector.is_changed(1, 1))
        self.assertTrue(detector.is_changed(2, 2))

    def test_line_expansion(self):
        detector = _TextChangeDetector("1\n2\n", "1\n3\n4\n2\n")
        self.assertFalse(detector.is_changed(1, 1))
        self.assertFalse(detector.is_changed(2, 2))

    def test_line_removals(self):
        detector = _TextChangeDetector("1\n3\n4\n2\n", "1\n2\n")
        self.assertFalse(detector.is_changed(1, 1))
        self.assertTrue(detector.is_changed(2, 3))
        self.assertFalse(detector.is_changed(4, 4))

    def test_multi_line_checks(self):
        detector = _TextChangeDetector("1\n2\n", "1\n3\n")
        self.assertTrue(detector.is_changed(1, 2))

    def test_consume_change(self):
        detector = _TextChangeDetector("1\n2\n", "1\n3\n")
        self.assertTrue(detector.is_changed(1, 2))
        self.assertTrue(detector.consume_changes(1, 2))
        self.assertFalse(detector.is_changed(1, 2))


class PyCoreProjectConfigsTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = None

    def tearDown(self):
        if self.project:
            testutils.remove_project(self.project)
        super().tearDown()

    def test_python_files_config(self):
        self.project = testutils.sample_project(python_files=["myscript"])
        myscript = self.project.root.create_file("myscript")
        self.assertTrue(self.project.pycore.is_python_file(myscript))

    def test_ignore_bad_imports(self):
        self.project = testutils.sample_project(ignore_bad_imports=True)
        code = "import some_nonexistent_module\n"
        pymod = libutils.get_string_module(self.project, code)
        self.assertFalse("some_nonexistent_module" in pymod)

    def test_ignore_bad_imports_for_froms(self):
        self.project = testutils.sample_project(ignore_bad_imports=True)
        code = "from some_nonexistent_module import var\n"
        pymod = libutils.get_string_module(self.project, code)
        self.assertFalse("var" in pymod)

    def test_reporting_syntax_errors_with_force_errors(self):
        self.project = testutils.sample_project(ignore_syntax_errors=True)
        mod = testutils.create_module(self.project, "mod")
        mod.write("syntax error ...\n")
        with self.assertRaises(exceptions.ModuleSyntaxError):
            self.project.pycore.resource_to_pyobject(mod, force_errors=True)

    def test_reporting_syntax_errors_in_strings_with_force_errors(self):
        self.project = testutils.sample_project(ignore_syntax_errors=True)
        with self.assertRaises(exceptions.ModuleSyntaxError):
            libutils.get_string_module(
                self.project, "syntax error ...", force_errors=True
            )

    def test_not_raising_errors_for_strings_with_ignore_errors(self):
        self.project = testutils.sample_project(ignore_syntax_errors=True)
        libutils.get_string_module(self.project, "syntax error ...")

    def test_reporting_syntax_errors_with_force_errors_for_packages(self):
        self.project = testutils.sample_project(ignore_syntax_errors=True)
        pkg = testutils.create_package(self.project, "pkg")
        pkg.get_child("__init__.py").write("syntax error ...\n")
        with self.assertRaises(exceptions.ModuleSyntaxError):
            self.project.pycore.resource_to_pyobject(pkg, force_errors=True)
